from __future__ import annotations

from datetime import datetime
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict


class SnapshotNode(BaseModel):
    model_config = ConfigDict(extra="allow")

    node_id: str
    type: str | None = None
    properties: dict[str, Any] | None = None
    confidence: float | None = None
    provenance: list[Any] | None = None


class SnapshotEdge(BaseModel):
    model_config = ConfigDict(extra="allow")

    edge_id: str | None = None
    from_node: str | None = None
    to_node: str | None = None
    type: str | None = None
    confidence: float | None = None
    provenance: Any | None = None


class Snapshot(BaseModel):
    model_config = ConfigDict(extra="allow")

    snapshot_id: str
    status: Literal["ready", "pending"] | None = None
    generated_at: datetime | None = None
    retrieval_time: datetime | None = None
    anchors: list[str] | None = None
    depth: int | None = None
    valid_time: datetime | None = None
    estimated_nodes: int | None = None

    nodes: list[SnapshotNode] | None = None
    edges: list[SnapshotEdge] | None = None
    decision_traces: list[dict[str, Any]] | None = None
    metadata: dict[str, Any] | None = None

    @classmethod
    def from_core(cls, core_obj: Any) -> Snapshot:
        if hasattr(core_obj, "to_dict") and callable(getattr(core_obj, "to_dict")):
            return cls.model_validate(core_obj.to_dict())
        if isinstance(core_obj, dict):
            return cls.model_validate(core_obj)
        raise TypeError("unsupported core object for Snapshot.from_core")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from sentinos_core import AuthenticatedClient, Client

from ._http import request_json, request_json_async, require_tenant


@dataclass
class BillingClient:
    _core: Client | AuthenticatedClient
    tenant_id: str | None = None

    def _tenant(self, tenant_id: str | None) -> str:
        return require_tenant(self.tenant_id, tenant_id)

    def get_entitlements(self, *, tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(self._core, "GET", "/v1/kernel/billing/entitlements", tenant_id=self._tenant(tenant_id))

    def patch_entitlements(self, *, patch: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "PATCH", "/v1/kernel/billing/entitlements", tenant_id=self._tenant(tenant_id), body=patch
        )

    async def get_entitlements_async(self, *, tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core, "GET", "/v1/kernel/billing/entitlements", tenant_id=self._tenant(tenant_id)
        )

    async def patch_entitlements_async(self, *, patch: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "PATCH",
            "/v1/kernel/billing/entitlements",
            tenant_id=self._tenant(tenant_id),
            body=patch,
        )

    def get_usage_summary(self, *, tenant_id: str | None = None, period_at: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core,
            "GET",
            "/v1/kernel/billing/usage/summary",
            tenant_id=self._tenant(tenant_id),
            params={"period_at": period_at},
        )

    async def get_usage_summary_async(
        self, *, tenant_id: str | None = None, period_at: str | None = None
    ) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "GET",
            "/v1/kernel/billing/usage/summary",
            tenant_id=self._tenant(tenant_id),
            params={"period_at": period_at},
        )

    def list_usage_events(
        self,
        *,
        tenant_id: str | None = None,
        metric_key: str | None = None,
        result: str | None = None,
        limit: int | None = None,
    ) -> dict[str, Any]:
        return request_json(
            self._core,
            "GET",
            "/v1/kernel/billing/usage/events",
            tenant_id=self._tenant(tenant_id),
            params={"metric_key": metric_key, "result": result, "limit": limit},
        )

    async def list_usage_events_async(
        self,
        *,
        tenant_id: str | None = None,
        metric_key: str | None = None,
        result: str | None = None,
        limit: int | None = None,
    ) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "GET",
            "/v1/kernel/billing/usage/events",
            tenant_id=self._tenant(tenant_id),
            params={"metric_key": metric_key, "result": result, "limit": limit},
        )



from __future__ import annotations

import asyncio
import json
from dataclasses import dataclass, field
from typing import Any

from sentinos.arbiter import ArbiterClient
from sentinos.chronos import ChronosClient
from sentinos.kernel import KernelClient
from sentinos.marketplace import MarketplaceClient
from sentinos.traces import TracesClient


@dataclass
class FakeResponse:
    payload: dict[str, Any]
    status_code: int = 200
    raw_content: bytes | None = None
    headers: dict[str, str] = field(default_factory=dict)

    def raise_for_status(self) -> None:
        if self.status_code >= 400:
            raise RuntimeError(f"http {self.status_code}")

    def json(self) -> dict[str, Any]:
        return self.payload

    @property
    def content(self) -> bytes:
        if self.raw_content is not None:
            return self.raw_content
        return json.dumps(self.payload).encode("utf-8")

    def iter_lines(self) -> list[str]:
        lines = self.payload.get("lines") or []
        return [str(x) for x in lines]


class FakeStreamResponse(FakeResponse):
    def __enter__(self) -> FakeStreamResponse:
        return self

    def __exit__(self, exc_type, exc, tb) -> bool:
        return False


class FakeHTTP:
    def __init__(self) -> None:
        self.calls: list[tuple[str, str, dict[str, Any] | None, dict[str, Any] | None]] = []

    def get(self, path: str, params: dict[str, Any] | None = None) -> FakeResponse:
        self.calls.append(("GET", path, params, None))
        if path == "/v1/trace/search":
            return FakeResponse({"next_cursor": "", "traces": None})
        if path.startswith("/v1/traces/") and path.endswith("/lineage"):
            return FakeResponse(
                {
                    "trace_id": "11111111-1111-1111-1111-111111111111",
                    "summary": {
                        "artifact_count": 4,
                        "side_effect_count": 3,
                        "blocked_count": 1,
                        "kinds": {"file": 1, "connector": 1, "domain": 1, "handoff": 1},
                        "top_domains": ["api.stripe.com"],
                        "top_connectors": ["stripe.refund"],
                        "top_outputs": ["refund.result"],
                        "has_handoff": True,
                        "has_writes": True,
                        "has_blocked_side_effects": True,
                    },
                    "artifacts": [
                        {
                            "artifact_id": "art-1",
                            "kind": "file",
                            "label": "refund_request.json",
                            "locator": "s3://acme/refund_request.json",
                            "status": "consumed",
                            "chronos_anchor": "file:refund_request.json",
                        },
                        {
                            "artifact_id": "art-2",
                            "kind": "domain",
                            "label": "api.stripe.com",
                            "locator": "https://api.stripe.com/v1/refunds",
                            "status": "blocked",
                            "metadata": {"related_trace_id": "22222222-2222-2222-2222-222222222222"},
                        },
                    ],
                    "events": [
                        {
                            "event_id": "evt-lineage-1",
                            "artifact_id": "art-1",
                            "action": "read",
                            "actor": "finance_bot",
                            "tool": "stripe.refund",
                            "timestamp": "2026-01-01T00:00:00Z",
                        },
                        {
                            "event_id": "evt-lineage-2",
                            "artifact_id": "art-2",
                            "action": "blocked",
                            "actor": "finance_bot",
                            "tool": "stripe.refund",
                            "timestamp": "2026-01-01T00:00:01Z",
                        },
                    ],
                }
            )
        if path == "/v1/trace/observability/distributed":
            return FakeResponse(
                {
                    "traces": [
                        {
                            "trace_id": "11111111-1111-1111-1111-111111111111",
                            "tenant_id": "acme",
                            "distributed_trace_id": "dd-trace-1",
                            "created_at": "2026-01-01T00:00:00Z",
                            "agent_id": "agent-1",
                            "policy_id": "policy-1",
                            "decision": "ALLOW",
                            "distributed_span_id": "span-1",
                        }
                    ]
                }
            )
        if path.startswith("/v1/trace/") and path.endswith("/verify"):
            return FakeResponse({"ok": True})
        if path.startswith("/v1/trace/") and path.endswith("/ledger"):
            return FakeResponse(
                {
                    "trace_id": "11111111-1111-1111-1111-111111111111",
                    "tenant_id": "acme",
                    "canonical_match": True,
                    "chain_match": True,
                    "chain_continuity_ok": True,
                    "verified": True,
                    "reason": "ok",
                }
            )
        if path == "/v1/trace/retention":
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "trace_days": 30,
                    "export_days": 90,
                    "ledger_days": 365,
                    "updated_at": "2026-01-01T00:00:00Z",
                }
            )
        if path.startswith("/v1/trace/export/job/"):
            return FakeResponse({"job_id": "j1", "status": "done"})
        if path == "/v1/arbitr/policies":
            return FakeResponse({"policies": [{"policy_id": "p1"}]})
        if path == "/v1/arbitr/policies/active":
            return FakeResponse({"keys": ["p1@v1"]})
        if path.startswith("/v1/arbitr/tenants/"):
            return FakeResponse({"tenant_id": "acme", "config": {}})
        if path.startswith("/v1/arbitr/simulate/"):
            return FakeResponse({"job_id": "sim-1", "status": "done"})
        if "/bundle" in path:
            return FakeResponse({"bundle": "ok"}, raw_content=b"bundle-bytes")
        if path == "/v1/arbitr/governance/dashboard":
            return FakeResponse({"violations_24h": 0})
        if path == "/v1/arbitr/governance/violations":
            return FakeResponse({"violations": [{"trace_id": "t1"}]})
        if path == "/v1/chronos/observability/traces":
            return FakeResponse({"traces": []})
        if path == "/v1/chronos/observability/anomalies":
            return FakeResponse({"anomalies": []})
        if path == "/v1/chronos/observability/anomalies/stats":
            return FakeResponse({"count": 0})
        if path == "/v1/chronos/observability/compliance":
            return FakeResponse({"score": 1.0})
        if path == "/v1/chronos/observability/patterns":
            return FakeResponse({"patterns": []})
        if path == "/v1/chronos/snapshot":
            return FakeResponse({"snapshot_id": "s1", "nodes": [], "edges": []})
        if path.startswith("/v1/chronos/ingest/"):
            return FakeResponse(
                {
                    "ingest": {
                        "ingest_id": "11111111-1111-1111-1111-111111111111",
                        "tenant_id": "acme",
                        "body_hash": "abc123",
                        "status": "accepted",
                        "created_at": "2026-01-01T00:00:00Z",
                        "updated_at": "2026-01-01T00:00:01Z",
                    }
                }
            )
        if path == "/v1/chronos/observability/entities/e1/risk":
            return FakeResponse({"entity_id": "e1", "risk_score": 0.0})
        if path == "/v1/integrations/channels":
            return FakeResponse({"channels": []})
        if path == "/v1/marketplace/packs":
            return FakeResponse(
                {
                    "packs": [
                        {
                            "pack_id": "pack-1",
                            "tenant_id": "acme",
                            "name": "Starter Pack",
                            "description": "Starter",
                            "author": "sentinos",
                            "version": "1.0.0",
                            "verified": True,
                            "tags": ["starter"],
                            "policies": [],
                            "created_at": "2026-01-01T00:00:00Z",
                        }
                    ]
                }
            )
        if path == "/v1/marketplace/installs":
            return FakeResponse({"installs": []})
        if path.startswith("/v1/marketplace/packs/"):
            return FakeResponse(
                {
                    "pack_id": "pack-1",
                    "tenant_id": "acme",
                    "name": "Starter Pack",
                    "description": "Starter",
                    "author": "sentinos",
                    "version": "1.0.0",
                    "verified": True,
                    "tags": ["starter"],
                    "policies": [],
                    "created_at": "2026-01-01T00:00:00Z",
                }
            )
        if path == "/v1/integrations/health":
            return FakeResponse({"summary": {"healthy": 1}})
        if path == "/v1/integrations/otel/config":
            return FakeResponse(
                {
                    "enabled": True,
                    "endpoint": "https://collector.example.com",
                    "protocol": "http/protobuf",
                    "traces_enabled": True,
                    "metrics_enabled": True,
                    "include_sentinos_extensions": True,
                    "include_internal_service_spans": False,
                    "resource_attributes": {"service.name": "sentinos"},
                    "header_keys_masked": ["authorization"],
                    "deep_link_template": "https://observability.example.com/trace/{trace_id}",
                    "privacy_mode": "policy_enforced",
                }
            )
        if path == "/v1/integrations/otel/status":
            return FakeResponse(
                {
                    "enabled": True,
                    "last_successful_export_at": "2026-01-01T00:00:00Z",
                    "queue_depth": 2,
                    "dropped_batch_count": 1,
                    "traces_exported": 14,
                    "metrics_exported": 14,
                }
            )
        if path == "/v1/compliance/reports/soc2":
            return FakeResponse({"tenant_id": "acme", "metrics": {}})
        if path == "/v1/compliance/evidence/controls":
            framework = str((params or {}).get("framework") or "SOC2").upper()
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "framework": framework,
                    "summary": {"controls_total": 2, "controls_passing": 2, "controls_attention": 0},
                    "controls": [],
                    "bundle_sha256": "abc123",
                }
            )
        if path == "/v1/kernel/cost/summary":
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "from": "2026-01-01T00:00:00Z",
                    "to": "2026-01-02T00:00:00Z",
                    "group_by": (params or {}).get("group_by") or "day",
                    "rows": [
                        {
                            "label": "openai/gpt-4.1-mini",
                            "total_usd": 12.4,
                            "reported_total_usd": 10.1,
                            "estimated_total_usd": 2.3,
                            "pricing_source": "mixed",
                            "total_tokens": 420000,
                            "blocked_cost_avoided_usd": 1.2,
                            "blocked_token_avoided": 8000,
                            "trace_count": 23,
                        }
                    ],
                }
            )
        if path == "/v1/kernel/cost/events":
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "from": "2026-01-01T00:00:00Z",
                    "to": "2026-01-02T00:00:00Z",
                    "events": [
                        {
                            "event_id": "evt-1",
                            "kind": "llm",
                            "label": "Primary model call",
                            "provider": "openai",
                            "model": "gpt-4.1-mini",
                            "total_tokens": 1420,
                            "reported_usd": 0.031,
                            "pricing_source": "reported",
                        }
                    ],
                }
            )
        if path == "/v1/kernel/cost/avoided":
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "from": "2026-01-01T00:00:00Z",
                    "to": "2026-01-02T00:00:00Z",
                    "rows": [
                        {
                            "label": "budget_guardrail",
                            "total_usd": 1.2,
                            "total_tokens": 8000,
                            "event_count": 3,
                        }
                    ],
                }
            )
        if path == "/v1/kernel/cost/anomalies":
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "from": "2026-01-01T00:00:00Z",
                    "to": "2026-01-02T00:00:00Z",
                    "anomalies": [
                        {
                            "anomaly_id": "anom-1",
                            "kind": "model_spike",
                            "severity": "high",
                            "dimension": "openai/gpt-4.1-mini",
                            "baseline": 1.4,
                            "observed": 4.7,
                            "delta": 3.3,
                            "trace_id": "11111111-1111-1111-1111-111111111111",
                            "detected_at": "2026-01-01T00:00:00Z",
                        }
                    ],
                }
            )
        return FakeResponse({})

    def post(self, path: str, json: dict[str, Any] | None = None) -> FakeResponse:
        self.calls.append(("POST", path, None, json))
        if path == "/v1/trace/export":
            return FakeResponse({"job_id": "j1", "status": "queued"})
        if path.startswith("/v1/trace/") and path.endswith("/replay"):
            return FakeResponse(
                {
                    "trace_id": "11111111-1111-1111-1111-111111111111",
                    "tenant_id": "acme",
                    "replayed_at": "2026-01-01T00:00:00Z",
                    "profile": (json or {}).get("profile", "active_policy_chain"),
                    "policy_source": "original_trace_policy",
                    "snapshot_source": "original_snapshot",
                    "fidelity": "deterministic",
                    "fidelity_reasons": ["Original policy metadata and snapshot were available."],
                    "reconstruction_basis": {
                        "policy_keys": ["sentinos/demo@v1"],
                        "original_policy_key": "sentinos/demo@v1",
                        "original_policy_id": "sentinos/demo@v1",
                        "original_policy_version": "v1",
                        "snapshot_id": "snap-1",
                    },
                    "evidence_export_ready": True,
                    "evidence_export_hints": ["Replay can be exported with ledger and privacy snapshots."],
                    "policy_keys": ["sentinos/demo@v1"],
                    "drift_detected": False,
                    "original": {
                        "decision": "ALLOW",
                        "policy_id": "sentinos/demo@v1",
                        "cost_breakdown": {
                            "total_usd": 0.031,
                            "reported_total_usd": 0.031,
                            "pricing_source": "reported",
                            "provider": "openai",
                            "model": "gpt-4.1-mini",
                            "total_tokens": 1420,
                        },
                        "cost_events": [
                            {
                                "event_id": "evt-1",
                                "kind": "llm",
                                "label": "Primary model call",
                                "provider": "openai",
                                "model": "gpt-4.1-mini",
                                "total_tokens": 1420,
                                "reported_usd": 0.031,
                                "pricing_source": "reported",
                            }
                        ],
                        "checks": [
                            {
                                "key": "tool.stripe_refund",
                                "label": "Tool access: stripe.refund",
                                "category": "tool",
                                "status": "ALLOWED",
                                "reason": "Allowed for finance workflow",
                                "matched": True,
                            }
                        ],
                    },
                    "replay": {
                        "decision": "ALLOW",
                        "policy_id": "sentinos/demo@v1",
                        "cost_breakdown": {
                            "total_usd": 0.044,
                            "estimated_total_usd": 0.044,
                            "pricing_source": "estimated",
                            "provider": "openai",
                            "model": "gpt-4.1-mini",
                            "total_tokens": 1800,
                        },
                        "cost_events": [
                            {
                                "event_id": "evt-2",
                                "kind": "retry",
                                "label": "Replay retry",
                                "provider": "openai",
                                "model": "gpt-4.1-mini",
                                "retry_index": 1,
                                "total_tokens": 1800,
                                "estimated_usd": 0.044,
                                "pricing_source": "estimated",
                            }
                        ],
                        "checks": [
                            {
                                "key": "tool.stripe_refund",
                                "label": "Tool access: stripe.refund",
                                "category": "tool",
                                "status": "ALLOWED",
                                "reason": "Allowed for finance workflow",
                                "matched": True,
                            }
                        ],
                    },
                    "comparison": {
                        "decision_changed": False,
                        "policy_changed": False,
                        "reason_changed": False,
                        "checks_added": 0,
                        "checks_removed": 0,
                        "checks_changed": 0,
                        "cost_changed": True,
                        "total_cost_delta_usd": 0.013,
                        "avoided_cost_delta_usd": 0.0,
                    },
                }
            )
        if path.startswith("/v1/trace/") and path.endswith("/replay/matrix"):
            return FakeResponse(
                {
                    "trace_id": "11111111-1111-1111-1111-111111111111",
                    "tenant_id": "acme",
                    "generated_at": "2026-01-01T00:00:00Z",
                    "entries": [
                        {
                            "profile": "active_policy_chain",
                            "response": {
                                "trace_id": "11111111-1111-1111-1111-111111111111",
                                "tenant_id": "acme",
                                "replayed_at": "2026-01-01T00:00:00Z",
                                "profile": "active_policy_chain",
                                "fidelity": "best_effort",
                                "fidelity_reasons": [
                                    "Original snapshot was unavailable; current policy chain was used."
                                ],
                                "drift_detected": True,
                                "replay": {"decision": "ESCALATE"},
                            },
                        },
                        {
                            "profile": "original_policy_and_snapshot",
                            "response": {
                                "trace_id": "11111111-1111-1111-1111-111111111111",
                                "tenant_id": "acme",
                                "replayed_at": "2026-01-01T00:00:00Z",
                                "profile": "original_policy_and_snapshot",
                                "fidelity": "deterministic",
                                "fidelity_reasons": ["Original policy metadata and snapshot were available."],
                                "drift_detected": False,
                                "replay": {"decision": "ALLOW"},
                            },
                        },
                    ],
                }
            )
        if path.startswith("/v1/trace/") and path.endswith("/replay/export"):
            return FakeResponse(
                {
                    "trace_id": "11111111-1111-1111-1111-111111111111",
                    "tenant_id": "acme",
                    "profile": (json or {}).get("profile", "original_policy_and_snapshot"),
                    "export_job": {
                        "job_id": "job-replay-1",
                        "payload_sha256": "sha256:abc123",
                        "retention_until": "2026-04-01T00:00:00Z",
                    },
                    "replay": {
                        "trace_id": "11111111-1111-1111-1111-111111111111",
                        "tenant_id": "acme",
                        "replayed_at": "2026-01-01T00:00:00Z",
                        "profile": (json or {}).get("profile", "original_policy_and_snapshot"),
                        "policy_source": "original_trace_policy",
                        "snapshot_source": "original_snapshot",
                        "fidelity": "deterministic",
                        "fidelity_reasons": ["Original policy metadata and snapshot were available."],
                        "evidence_export_ready": True,
                        "policy_keys": ["sentinos/demo@v1"],
                        "drift_detected": False,
                        "replay": {"decision": "ALLOW"},
                    },
                }
            )
        if path == "/v1/trace/retention/enforce":
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "dry_run": bool((json or {}).get("dry_run", False)),
                    "traces_cutoff": "2025-01-01T00:00:00Z",
                    "exports_cutoff": "2025-01-01T00:00:00Z",
                    "ledger_cutoff": "2025-01-01T00:00:00Z",
                    "traces_affected": 12,
                    "exports_affected": 3,
                    "ledger_anchors_affected": 0,
                }
            )
        if path == "/v1/chronos/query":
            return FakeResponse(
                {
                    "query_id": "q1",
                    "generated_at": "2026-01-01T00:00:00Z",
                    "hits": None,
                    "ranking": None,
                    "nodes": None,
                    "edges": None,
                }
            )
        if path == "/v1/arbitr/governance/reports":
            return FakeResponse({"report_id": "r1"})
        if path in ("/v1/arbitr/evaluate", "/v1/arbitr/compile"):
            return FakeResponse({"ok": True})
        if path == "/v1/chronos/observability/traces/analyze":
            return FakeResponse({"score": 0.42})
        if path == "/v1/chronos/observability/compliance/report":
            return FakeResponse({"report_id": "c1"})
        if path == "/v1/chronos/ingest":
            return FakeResponse({"ok": True})
        if path == "/v1/reprocess":
            return FakeResponse({"accepted": 2, "reprocessed": 2})
        if path == "/v1/integrations/channels":
            return FakeResponse({"channel_id": "ch1", "kind": "SLACK", "name": "ops"}, status_code=201)
        if path.endswith("/install") and "/v1/marketplace/packs/" in path:
            return FakeResponse({"install_id": "inst-1", "simulation_job_ids": []})
        if path == "/v1/integrations/channels/validate":
            return FakeResponse({"ok": True})
        if path == "/v1/integrations/otel/test":
            return FakeResponse(
                {
                    "ok": True,
                    "message": "collector responded with 202",
                    "trace_delivered": True,
                    "metrics_delivered": True,
                }
            )
        if path == "/v1/integrations/datadog/export":
            return FakeResponse({"exported_metrics": 12})
        if path == "/v1/integrations/siem/export":
            return FakeResponse({"exported_records": 7})
        if path.endswith("/test"):
            return FakeResponse({"success": True})
        return FakeResponse({})

    def put(self, path: str, json: dict[str, Any] | None = None) -> FakeResponse:
        self.calls.append(("PUT", path, None, json))
        if path == "/v1/integrations/otel/config":
            payload = dict(json or {})
            payload.setdefault("header_keys_masked", sorted((payload.get("header_values_write_only") or {}).keys()))
            return FakeResponse(payload)
        if path.startswith("/v1/arbitr/tenants/"):
            return FakeResponse({"tenant_id": "acme", "config": json or {}})
        return FakeResponse({"channel_id": "ch1", "enabled": True})

    def patch(self, path: str, json: dict[str, Any] | None = None) -> FakeResponse:
        self.calls.append(("PATCH", path, None, json))
        if path == "/v1/trace/retention":
            trace_days = int((json or {}).get("trace_days", 30))
            export_days = int((json or {}).get("export_days", 90))
            ledger_days = int((json or {}).get("ledger_days", 365))
            return FakeResponse(
                {
                    "tenant_id": "acme",
                    "trace_days": trace_days,
                    "export_days": export_days,
                    "ledger_days": ledger_days,
                    "updated_at": "2026-01-01T00:00:01Z",
                }
            )
        return FakeResponse({})

    def delete(self, path: str) -> FakeResponse:
        self.calls.append(("DELETE", path, None, None))
        return FakeResponse({"ok": True})

    def request(
        self,
        method: str,
        url: str,
        params: dict[str, Any] | None = None,
        json: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
    ) -> FakeResponse:
        _ = headers
        method = method.upper()
        if method == "GET":
            return self.get(url, params=params)
        if method == "POST":
            return self.post(url, json=json)
        if method == "PUT":
            return self.put(url, json=json)
        if method == "PATCH":
            return self.patch(url, json=json)
        if method == "DELETE":
            return self.delete(url)
        raise RuntimeError(f"unsupported method {method}")

    def stream(
        self,
        method: str,
        path: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        timeout: float | None = None,
    ) -> FakeStreamResponse:
        _ = headers, timeout
        self.calls.append((method, path, params, None))
        return FakeStreamResponse({"lines": ["event: alert", "data: {}"]})


class FakeAsyncHTTP:
    def __init__(self, http: FakeHTTP) -> None:
        self._http = http

    async def get(self, path: str, params: dict[str, Any] | None = None) -> FakeResponse:
        return self._http.get(path, params=params)

    async def post(self, path: str, json: dict[str, Any] | None = None) -> FakeResponse:
        return self._http.post(path, json=json)

    async def put(self, path: str, json: dict[str, Any] | None = None) -> FakeResponse:
        return self._http.put(path, json=json)

    async def patch(self, path: str, json: dict[str, Any] | None = None) -> FakeResponse:
        return self._http.patch(path, json=json)

    async def delete(self, path: str) -> FakeResponse:
        return self._http.delete(path)

    async def request(
        self,
        method: str,
        url: str,
        params: dict[str, Any] | None = None,
        json: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
    ) -> FakeResponse:
        _ = headers
        method = method.upper()
        if method == "GET":
            return self._http.get(url, params=params)
        if method == "POST":
            return self._http.post(url, json=json)
        if method == "PUT":
            return self._http.put(url, json=json)
        if method == "PATCH":
            return self._http.patch(url, json=json)
        if method == "DELETE":
            return self._http.delete(url)
        raise RuntimeError(f"unsupported method {method}")


class FakeCore:
    def __init__(self, http: FakeHTTP, headers: dict[str, str] | None = None) -> None:
        self._http = http
        self._headers = headers or {}

    def with_headers(self, headers: dict[str, str]) -> FakeCore:
        merged = dict(self._headers)
        merged.update(headers)
        return FakeCore(self._http, merged)

    def get_httpx_client(self) -> FakeHTTP:
        return self._http

    def get_async_httpx_client(self) -> FakeAsyncHTTP:
        return FakeAsyncHTTP(self._http)


def test_traces_handles_null_list_payload() -> None:
    c = TracesClient(FakeCore(FakeHTTP()), tenant_id="acme")
    traces = c.list_traces()
    assert traces == []


def test_chronos_query_normalizes_nullable_arrays() -> None:
    c = ChronosClient(FakeCore(FakeHTTP()), tenant_id="acme")
    out = c.query(query="test", limit=5)
    assert out.hits == []
    assert out.nodes == []
    assert out.edges == []
    assert out.additional_properties.get("ranking") == []


def test_arbiter_governance_endpoints() -> None:
    c = ArbiterClient(FakeCore(FakeHTTP()), tenant_id="acme")
    dashboard = c.governance_dashboard()
    violations = c.governance_violations(limit=100)
    report = c.governance_report(limit=200)

    assert dashboard["violations_24h"] == 0
    assert len(violations) == 1
    assert report["report_id"] == "r1"


def test_kernel_integrations_compliance_and_sse() -> None:
    c = KernelClient(FakeCore(FakeHTTP()), tenant_id="acme")

    created = c.create_notification_channel(kind="SLACK", name="ops", config={"webhook_url": "https://x"})
    listed = c.list_notification_channels()
    tested = c.test_notification_channel("ch1", message="ping")
    valid = c.validate_notification_channel_config(kind="SLACK", config={"webhook_url": "https://x"})
    health = c.integrations_health()
    dd = c.export_datadog(api_key="dd", app_key="app")
    siem = c.export_siem(export_format="JSON")
    soc2 = c.soc2_report()
    controls = c.compliance_control_evidence_report(framework="FEDRAMP")
    stream_lines = list(c.events_stream())

    assert created["channel_id"] == "ch1"
    assert listed == []
    assert tested["success"] is True
    assert valid is True
    assert health["summary"]["healthy"] == 1
    assert dd["exported_metrics"] == 12
    assert siem["exported_records"] == 7
    assert soc2["tenant_id"] == "acme"
    assert controls["framework"] == "FEDRAMP"
    assert stream_lines == ["event: alert", "data: {}"]


def test_kernel_cost_observability_methods() -> None:
    c = KernelClient(FakeCore(FakeHTTP()), tenant_id="acme")

    summary = c.get_cost_summary(group_by="provider_model")
    events = c.list_cost_events(provider="openai", limit=5)
    avoided = c.get_cost_avoided()
    anomalies = c.list_cost_anomalies()

    assert summary.group_by == "provider_model"
    assert summary.rows[0].pricing_source == "mixed"
    assert events.events[0].kind == "llm"
    assert events.events[0].reported_usd == 0.031
    assert avoided.rows[0].label == "budget_guardrail"
    assert anomalies.anomalies[0].kind == "model_spike"


def test_kernel_otel_bridge_methods() -> None:
    c = KernelClient(FakeCore(FakeHTTP()), tenant_id="acme")

    config = c.get_otel_export_config()
    updated = c.update_otel_export_config(
        config={
            "enabled": True,
            "endpoint": "https://collector.example.com",
            "protocol": "http/protobuf",
            "traces_enabled": True,
            "metrics_enabled": True,
            "include_sentinos_extensions": True,
            "include_internal_service_spans": False,
            "resource_attributes": {"service.name": "sentinos"},
            "header_values_write_only": {"authorization": "Bearer token"},
            "privacy_mode": "policy_enforced",
        }
    )
    status = c.get_otel_export_status()
    tested = c.test_otel_export()

    assert config.enabled is True
    assert updated.header_keys_masked == ["authorization"]
    assert status.queue_depth == 2
    assert tested.ok is True
    assert tested.trace_delivered is True


def test_kernel_integrations_compliance_and_sse_async() -> None:
    c = KernelClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        listed = await c.list_notification_channels_async()
        created = await c.create_notification_channel_async(
            kind="SLACK", name="ops", config={"webhook_url": "https://x"}
        )
        tested = await c.test_notification_channel_async("ch1", message="ping")
        valid = await c.validate_notification_channel_config_async(kind="SLACK", config={"webhook_url": "https://x"})
        health = await c.integrations_health_async()
        dd = await c.export_datadog_async(api_key="dd")
        siem = await c.export_siem_async()
        soc2 = await c.soc2_report_async()
        controls = await c.compliance_control_evidence_report_async(framework="HIPAA")

        assert listed == []
        assert created["channel_id"] == "ch1"
        assert tested["success"] is True
        assert valid is True
        assert health["summary"]["healthy"] == 1
        assert dd["exported_metrics"] == 12
        assert siem["exported_records"] == 7
        assert soc2["tenant_id"] == "acme"
        assert controls["framework"] == "HIPAA"

    asyncio.run(run())


def test_kernel_cost_observability_methods_async() -> None:
    c = KernelClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        summary = await c.get_cost_summary_async(group_by="actor")
        events = await c.list_cost_events_async(actor="finance_bot")
        avoided = await c.get_cost_avoided_async()
        anomalies = await c.list_cost_anomalies_async()

        assert summary.group_by == "actor"
        assert summary.rows[0].total_tokens == 420000
        assert events.events[0].provider == "openai"
        assert avoided.rows[0].total_usd == 1.2
        assert anomalies.anomalies[0].severity == "high"

    asyncio.run(run())


def test_kernel_otel_bridge_methods_async() -> None:
    c = KernelClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        config = await c.get_otel_export_config_async()
        updated = await c.update_otel_export_config_async(
            config={
                "enabled": True,
                "endpoint": "https://collector.example.com",
                "protocol": "http/protobuf",
                "traces_enabled": True,
                "metrics_enabled": True,
                "include_sentinos_extensions": True,
                "include_internal_service_spans": False,
                "resource_attributes": {"service.name": "sentinos"},
                "header_values_write_only": {"x-api-key": "secret"},
                "privacy_mode": "policy_enforced",
            }
        )
        status = await c.get_otel_export_status_async()
        tested = await c.test_otel_export_async()

        assert config.deep_link_template == "https://observability.example.com/trace/{trace_id}"
        assert updated.header_keys_masked == ["x-api-key"]
        assert status.traces_exported == 14
        assert tested.metrics_delivered is True

    asyncio.run(run())


def test_arbiter_async_surface_methods() -> None:
    c = ArbiterClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        dashboard = await c.governance_dashboard_async()
        violations = await c.governance_violations_async(limit=10)
        report = await c.governance_report_async(limit=20)
        evaluate = await c.evaluate_async(request={"tenant_id": "acme"})
        compiled = await c.compile_async(request={"rego": "package x"})
        policies = await c.list_policies_async()
        active = await c.active_policies_async(tool="stripe.transfer")
        tenant = await c.get_tenant_config_async()
        updated = await c.upsert_tenant_config_async(config={"tenant_id": "acme"})
        bundle = await c.get_policy_bundle_async(policy_id="p1", version="v1")
        job = await c.get_simulation_job_async("sim-1")

        assert dashboard["violations_24h"] == 0
        assert len(violations) == 1
        assert report["report_id"] == "r1"
        assert evaluate["ok"] is True
        assert compiled["ok"] is True
        assert policies[0]["policy_id"] == "p1"
        assert active == ["p1@v1"]
        assert tenant["tenant_id"] == "acme"
        assert updated["tenant_id"] == "acme"
        assert bundle
        assert job["job_id"] == "sim-1"

    asyncio.run(run())


def test_chronos_async_surface_methods() -> None:
    c = ChronosClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        traces = await c.observability_traces_async()
        analyzed = await c.analyze_traces_async()
        anomalies = await c.observability_anomalies_async()
        stats = await c.observability_anomaly_stats_async()
        compliance = await c.observability_compliance_async()
        report = await c.observability_compliance_report_async()
        patterns = await c.observability_patterns_async()
        risk = await c.entity_risk_async("e1")
        ingest = await c.get_ingest_status_async("11111111-1111-1111-1111-111111111111")
        event = await c.ingest_event_async(event={"event_type": "trace.indexed"})
        reprocess = await c.reprocess_traces_async(request={"trace_ids": []})
        legacy = await c.get_snapshot_legacy_async(anchor="agent:foo")

        assert traces["traces"] == []
        assert analyzed["score"] == 0.42
        assert anomalies["anomalies"] == []
        assert stats["count"] == 0
        assert compliance["score"] == 1.0
        assert report["report_id"] == "c1"
        assert patterns["patterns"] == []
        assert risk["entity_id"] == "e1"
        assert ingest["ingest"]["status"] == "accepted"
        assert event["ok"] is True
        assert reprocess["reprocessed"] == 2
        assert legacy.snapshot_id == "s1"

    asyncio.run(run())


def test_traces_async_export_verify_and_status() -> None:
    c = TracesClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        verified = await c.verify_trace_async("11111111-1111-1111-1111-111111111111")
        exported = await c.export_traces_async(request={})
        status = await c.export_status_async("11111111-1111-1111-1111-111111111111")
        assert verified["ok"] is True
        assert exported["job_id"] == "j1"
        assert status["status"] == "done"

    asyncio.run(run())


def test_traces_export_omits_tenant_id_from_body() -> None:
    http = FakeHTTP()
    c = TracesClient(FakeCore(http), tenant_id="acme")

    c.export_traces(request={"limit": 10, "tenant_id": "wrong"})

    export_calls = [call for call in http.calls if call[0] == "POST" and call[1] == "/v1/trace/export"]
    assert len(export_calls) == 1
    _, _, _, payload = export_calls[0]
    assert isinstance(payload, dict)
    assert payload.get("limit") == 10
    assert "tenant_id" not in payload


def test_traces_replay_ledger_retention_and_distributed_methods() -> None:
    c = TracesClient(FakeCore(FakeHTTP()), tenant_id="acme")

    replayed = c.replay_trace(
        "11111111-1111-1111-1111-111111111111",
        request={
            "profile": "original_policy_and_snapshot",
            "include_explain": True,
            "include_evidence_hints": True,
        },
    )
    matrix = c.replay_trace_matrix(
        "11111111-1111-1111-1111-111111111111",
        request={"include_explain": True},
    )
    exported_replay = c.export_replay_evidence(
        "11111111-1111-1111-1111-111111111111",
        request={"profile": "original_policy_and_snapshot", "include_explain": True},
    )
    verified = c.ledger_verify("11111111-1111-1111-1111-111111111111")
    retention_before = c.get_retention_policy()
    retention_after = c.update_retention_policy(request={"trace_days": 45, "export_days": 120, "ledger_days": 730})
    enforced = c.enforce_retention(request={"dry_run": True})
    distributed = c.distributed_trace_summaries(limit=25)

    assert replayed.drift_detected is False
    assert replayed.tenant_id == "acme"
    assert replayed.profile == "original_policy_and_snapshot"
    assert replayed.fidelity == "deterministic"
    assert replayed.original.checks[0].label == "Tool access: stripe.refund"
    assert replayed.original.checks[0].status == "ALLOWED"
    assert replayed.comparison.checks_changed == 0
    assert replayed.original.cost_breakdown.total_usd == 0.031
    assert replayed.replay.cost_events[0].kind == "retry"
    assert replayed.comparison.cost_changed is True
    assert matrix.entries[0].profile == "active_policy_chain"
    assert matrix.entries[1].response.fidelity == "deterministic"
    assert exported_replay.export_job["job_id"] == "job-replay-1"
    assert exported_replay.replay.evidence_export_ready is True
    assert verified.verified is True
    assert retention_before.trace_days == 30
    assert retention_after.trace_days == 45
    assert enforced.dry_run is True
    assert enforced.traces_affected == 12
    assert len(distributed) == 1
    assert distributed[0].distributed_trace_id == "dd-trace-1"


def test_traces_lineage_methods() -> None:
    c = TracesClient(FakeCore(FakeHTTP()), tenant_id="acme")

    lineage = c.get_trace_lineage("11111111-1111-1111-1111-111111111111")

    assert lineage.summary.artifact_count == 4
    assert lineage.summary.top_domains == ["api.stripe.com"]
    assert lineage.artifacts[0].kind == "file"
    assert lineage.events[1].action == "blocked"


def test_traces_lineage_methods_async() -> None:
    c = TracesClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        lineage = await c.get_trace_lineage_async("11111111-1111-1111-1111-111111111111")
        assert lineage.summary.has_blocked_side_effects is True
        assert lineage.artifacts[1].status == "blocked"

    asyncio.run(run())


def test_traces_replay_ledger_retention_and_distributed_methods_async() -> None:
    c = TracesClient(FakeCore(FakeHTTP()), tenant_id="acme")

    async def run() -> None:
        replayed = await c.replay_trace_async(
            "11111111-1111-1111-1111-111111111111",
            request={
                "profile": "original_policy",
                "policy_keys": ["sentinos/demo@v1"],
            },
        )
        matrix = await c.replay_trace_matrix_async(
            "11111111-1111-1111-1111-111111111111",
            request={"include_explain": True},
        )
        exported_replay = await c.export_replay_evidence_async(
            "11111111-1111-1111-1111-111111111111",
            request={"profile": "original_policy_and_snapshot", "include_explain": True},
        )
        verified = await c.ledger_verify_async("11111111-1111-1111-1111-111111111111")
        retention_before = await c.get_retention_policy_async()
        retention_after = await c.update_retention_policy_async(
            request={"trace_days": 60, "export_days": 150, "ledger_days": 730}
        )
        enforced = await c.enforce_retention_async(request={"dry_run": False})
        distributed = await c.distributed_trace_summaries_async(limit=50)

        assert replayed.drift_detected is False
        assert replayed.profile == "original_policy"
        assert replayed.replay.checks[0].category == "tool"
        assert replayed.comparison.decision_changed is False
        assert replayed.replay.cost_breakdown.pricing_source == "estimated"
        assert replayed.comparison.total_cost_delta_usd == 0.013
        assert matrix.entries[0].response.fidelity == "best_effort"
        assert exported_replay.profile == "original_policy_and_snapshot"
        assert verified.chain_continuity_ok is True
        assert retention_before.trace_days == 30
        assert retention_after.trace_days == 60
        assert enforced.dry_run is False
        assert distributed[0].decision == "ALLOW"

    asyncio.run(run())


def test_marketplace_sync_and_async_methods() -> None:
    c = MarketplaceClient(FakeCore(FakeHTTP()), tenant_id="acme")

    packs = c.list_packs()
    fetched = c.get_pack(pack_id="pack-1")
    install = c.install_pack(pack_id="pack-1")
    installs = c.list_installs()
    c.uninstall_pack(pack_id="pack-1")

    assert packs[0].pack_id == "pack-1"
    assert fetched.pack_id == "pack-1"
    assert install.install_id == "inst-1"
    assert installs == []

    async def run() -> None:
        packs_async = await c.list_packs_async()
        fetched_async = await c.get_pack_async(pack_id="pack-1")
        install_async = await c.install_pack_async(pack_id="pack-1")
        installs_async = await c.list_installs_async()
        await c.uninstall_pack_async(pack_id="pack-1")

        assert packs_async[0].pack_id == "pack-1"
        assert fetched_async.pack_id == "pack-1"
        assert install_async.install_id == "inst-1"
        assert installs_async == []

    asyncio.run(run())



from .alert import Alert, AlertRule, Anomaly
from .api_key import APIKeyRecord
from .cost import (
    KernelCostAnomaliesResponse,
    KernelCostAnomaly,
    KernelCostAvoidedResponse,
    KernelCostAvoidedRow,
    KernelCostEventsResponse,
    KernelCostSummaryResponse,
    KernelCostSummaryRow,
    TraceCostActorBreakdown,
    TraceCostBreakdown,
    TraceCostEvent,
    TraceCostProviderModelBreakdown,
    TraceCostRetryBreakdown,
    TraceCostToolBreakdown,
)
from .decision_trace import (
    DecisionTrace,
    DecisionTracePolicyCheck,
    DecisionTracePolicyEvaluation,
)
from .incident import Incident, IncidentTimelineEvent
from .lineage import (
    TraceArtifactLineageEvent,
    TraceArtifactLineageResponse,
    TraceArtifactLineageSummary,
    TraceArtifactRef,
)
from .marketplace import InstallResult, MarketplacePack, PackInstall
from .openresponses import (
    OpenResponsesError,
    OpenResponsesItem,
    OpenResponsesRequest,
    OpenResponsesResponse,
)
from .otel import OtelExportConfig, OtelExportStatus, OtelExportTestResult
from .policy import PolicyMetadata
from .replay import (
    TraceReplayComparison,
    TraceReplayDecision,
    TraceReplayExportResponse,
    TraceReplayMatrixEntry,
    TraceReplayMatrixResponse,
    TraceReplayReconstructionBasis,
    TraceReplayResponse,
)
from .snapshot import Snapshot

__all__ = (
    "APIKeyRecord",
    "TraceCostBreakdown",
    "TraceCostEvent",
    "TraceCostProviderModelBreakdown",
    "TraceCostRetryBreakdown",
    "TraceCostToolBreakdown",
    "TraceCostActorBreakdown",
    "KernelCostSummaryRow",
    "KernelCostSummaryResponse",
    "KernelCostEventsResponse",
    "KernelCostAvoidedRow",
    "KernelCostAvoidedResponse",
    "KernelCostAnomaly",
    "KernelCostAnomaliesResponse",
    "DecisionTrace",
    "DecisionTracePolicyCheck",
    "DecisionTracePolicyEvaluation",
    "TraceReplayDecision",
    "TraceReplayComparison",
    "TraceReplayReconstructionBasis",
    "TraceReplayResponse",
    "TraceReplayMatrixEntry",
    "TraceReplayMatrixResponse",
    "TraceReplayExportResponse",
    "TraceArtifactLineageSummary",
    "TraceArtifactRef",
    "TraceArtifactLineageEvent",
    "TraceArtifactLineageResponse",
    "Alert",
    "AlertRule",
    "Anomaly",
    "Incident",
    "IncidentTimelineEvent",
    "InstallResult",
    "MarketplacePack",
    "PackInstall",
    "OpenResponsesRequest",
    "OpenResponsesResponse",
    "OpenResponsesItem",
    "OpenResponsesError",
    "OtelExportConfig",
    "OtelExportStatus",
    "OtelExportTestResult",
    "PolicyMetadata",
    "Snapshot",
)



"""Sentinos SDK public exports."""

from __future__ import annotations

from .a2a import A2AClient
from .alerts import AlertsClient
from .arbiter import ArbiterClient
from .audit import AuditClient
from .auth import (
    APIKeyAuth,
    JWTAuth,
    WorkforceAssertion,
    WorkforceMappingError,
    WorkforcePolicyDeniedError,
    WorkforceSessionRevokedError,
    WorkforceTokenError,
    WorkforceTokenProvider,
)
from .billing import BillingClient
from .chronos import ChronosClient
from .client import SentinosClient
from .controlplane import ControlplaneClient
from .dashboards import DashboardsClient
from .incidents import IncidentsClient
from .integrations import (
    AnthropicMessagesAdapter,
    AsyncAnthropicMessagesAdapter,
    AsyncBedrockConverseAdapter,
    AsyncOpenAIChatCompletionsAdapter,
    AsyncOpenAIResponsesAdapter,
    AsyncOpenResponsesAdapter,
    BedrockAccessDeniedError,
    BedrockConverseAdapter,
    BedrockConverseStreamResult,
    BedrockProviderError,
    BedrockThrottlingError,
    BedrockValidationError,
    LLMGuard,
    LLMPolicyDeniedError,
    LLMPolicyEscalationError,
    LLMPolicyExecutionError,
    LLMPolicyResult,
    OpenAIChatCompletionsAdapter,
    OpenAIResponsesAdapter,
    OpenResponsesAdapter,
    OpenResponsesStreamResult,
    create_anthropic_messages_adapter,
    create_async_anthropic_messages_adapter,
    create_async_bedrock_converse_adapter,
    create_async_openai_chat_adapter,
    create_async_openai_responses_adapter,
    create_async_openresponses_adapter,
    create_async_openrouter_chat_adapter,
    create_async_openrouter_openresponses_adapter,
    create_async_openrouter_responses_adapter,
    create_bedrock_converse_adapter,
    create_openai_chat_adapter,
    create_openai_responses_adapter,
    create_openresponses_adapter,
    create_openrouter_chat_adapter,
    create_openrouter_openresponses_adapter,
    create_openrouter_responses_adapter,
    guard_anthropic_messages,
    guard_openai_chat_completion,
    guard_openresponses_create,
    iter_openresponses_sse_lines,
    make_guarded_tool,
    make_guarded_tool_async,
    parse_openresponses_sse_event,
)
from .kernel import KernelClient
from .marketplace import MarketplaceClient
from .models.alert import Alert, AlertRule, Anomaly
from .models.cost import (
    KernelCostAnomaliesResponse,
    KernelCostAnomaly,
    KernelCostAvoidedResponse,
    KernelCostAvoidedRow,
    KernelCostEventsResponse,
    KernelCostSummaryResponse,
    KernelCostSummaryRow,
    TraceCostActorBreakdown,
    TraceCostBreakdown,
    TraceCostEvent,
    TraceCostProviderModelBreakdown,
    TraceCostRetryBreakdown,
    TraceCostToolBreakdown,
)
from .models.decision_trace import DecisionTrace, PolicyEvaluation
from .models.incident import Incident, IncidentTimelineEvent
from .models.openresponses import OpenResponsesError, OpenResponsesItem, OpenResponsesRequest, OpenResponsesResponse
from .models.otel import OtelExportConfig, OtelExportStatus, OtelExportTestResult
from .models.policy import PolicyItem, PolicyMetadata
from .models.replay import (
    TraceReplayComparison,
    TraceReplayDecision,
    TraceReplayExportResponse,
    TraceReplayMatrixEntry,
    TraceReplayMatrixResponse,
    TraceReplayReconstructionBasis,
    TraceReplayResponse,
)
from .privacy import PrivacyClient
from .traces import TracesClient
from .utils.telemetry import TelemetryConfig, httpx_event_hooks

__all__ = [
    "A2AClient",
    "APIKeyAuth",
    "Alert",
    "AlertsClient",
    "AlertRule",
    "Anomaly",
    "AuditClient",
    "ArbiterClient",
    "BillingClient",
    "ChronosClient",
    "ControlplaneClient",
    "DashboardsClient",
    "DecisionTrace",
    "TraceCostBreakdown",
    "TraceCostEvent",
    "TraceCostProviderModelBreakdown",
    "TraceCostRetryBreakdown",
    "TraceCostToolBreakdown",
    "TraceCostActorBreakdown",
    "KernelCostSummaryRow",
    "KernelCostSummaryResponse",
    "KernelCostEventsResponse",
    "KernelCostAvoidedRow",
    "KernelCostAvoidedResponse",
    "KernelCostAnomaly",
    "KernelCostAnomaliesResponse",
    "Incident",
    "IncidentsClient",
    "IncidentTimelineEvent",
    "JWTAuth",
    "WorkforceAssertion",
    "WorkforceTokenProvider",
    "WorkforceTokenError",
    "WorkforcePolicyDeniedError",
    "WorkforceMappingError",
    "WorkforceSessionRevokedError",
    "KernelClient",
    "LLMGuard",
    "BedrockConverseAdapter",
    "AsyncBedrockConverseAdapter",
    "BedrockConverseStreamResult",
    "BedrockProviderError",
    "BedrockAccessDeniedError",
    "BedrockThrottlingError",
    "BedrockValidationError",
    "OpenResponsesAdapter",
    "AsyncOpenResponsesAdapter",
    "OpenResponsesStreamResult",
    "OpenAIChatCompletionsAdapter",
    "AsyncOpenAIChatCompletionsAdapter",
    "OpenAIResponsesAdapter",
    "AsyncOpenAIResponsesAdapter",
    "AnthropicMessagesAdapter",
    "AsyncAnthropicMessagesAdapter",
    "create_openai_chat_adapter",
    "create_async_openai_chat_adapter",
    "create_openrouter_chat_adapter",
    "create_async_openrouter_chat_adapter",
    "create_openresponses_adapter",
    "create_async_openresponses_adapter",
    "create_openrouter_openresponses_adapter",
    "create_async_openrouter_openresponses_adapter",
    "create_openai_responses_adapter",
    "create_async_openai_responses_adapter",
    "create_openrouter_responses_adapter",
    "create_async_openrouter_responses_adapter",
    "create_anthropic_messages_adapter",
    "create_async_anthropic_messages_adapter",
    "create_bedrock_converse_adapter",
    "create_async_bedrock_converse_adapter",
    "LLMPolicyDeniedError",
    "LLMPolicyEscalationError",
    "LLMPolicyExecutionError",
    "LLMPolicyResult",
    "MarketplaceClient",
    "PolicyEvaluation",
    "TraceReplayDecision",
    "TraceReplayComparison",
    "TraceReplayReconstructionBasis",
    "TraceReplayResponse",
    "TraceReplayMatrixEntry",
    "TraceReplayMatrixResponse",
    "TraceReplayExportResponse",
    "PolicyItem",
    "PolicyMetadata",
    "SentinosClient",
    "TracesClient",
    "guard_anthropic_messages",
    "guard_openai_chat_completion",
    "guard_openresponses_create",
    "parse_openresponses_sse_event",
    "iter_openresponses_sse_lines",
    "make_guarded_tool",
    "make_guarded_tool_async",
    "OpenResponsesRequest",
    "OpenResponsesResponse",
    "OpenResponsesItem",
    "OpenResponsesError",
    "PrivacyClient",
    "TelemetryConfig",
    "httpx_event_hooks",
    "OtelExportConfig",
    "OtelExportStatus",
    "OtelExportTestResult",
]



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class TelemetryConfig:
    enabled: bool = True
    service_name: str = "sentinos-sdk"


def httpx_event_hooks(
    span_prefix: str,
    *,
    extra_attributes: dict[str, Any] | None = None,
) -> dict[str, list[Callable[..., Any]]]:
    """
    Return httpx event hooks for OpenTelemetry spans.

    Attach these via `httpx_args={"event_hooks": ...}` when constructing the generated core client.
    """
    extra_attributes = extra_attributes or {}

    def on_request(request: Any) -> None:
        if not extra_attributes:
            return
        try:
            request.extensions.setdefault("sentinos_attrs", {}).update(extra_attributes)
        except Exception:
            return

    def on_response(response: Any) -> None:
        # Best-effort: create span per request if opentelemetry is installed.
        try:
            from opentelemetry import trace as otel_trace
        except Exception:
            return
        tracer = otel_trace.get_tracer("sentinos-sdk")
        req = response.request
        name = f"{span_prefix} {req.method} {req.url.path}"
        with tracer.start_as_current_span(name) as span:
            span.set_attribute("http.method", req.method)
            span.set_attribute("http.url", str(req.url))
            span.set_attribute("http.status_code", response.status_code)
            attrs = req.extensions.get("sentinos_attrs") if hasattr(req, "extensions") else None
            if isinstance(attrs, dict):
                for k, v in attrs.items():
                    span.set_attribute(str(k), v)

    return {"request": [on_request], "response": [on_response]}



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import Any

import pytest

from sentinos.integrations import (
    LLMGuard,
    LLMPolicyDeniedError,
    LLMPolicyEscalationError,
    guard_openai_chat_completion,
)
from sentinos.models.decision_trace import DecisionTrace


def _trace(decision: str) -> DecisionTrace:
    return DecisionTrace.model_validate(
        {
            "trace_id": "trace-1",
            "timestamp": "2026-02-07T00:00:00Z",
            "tenant_id": "acme",
            "agent_id": "agent-1",
            "session_id": "sess-1",
            "intent": {"type": "llm_call", "tool": "llm.openai.chat.completions", "args": {}},
            "policy_evaluation": {
                "policy_id": "llm-guard",
                "policy_version": "v1",
                "decision": decision,
                "reason": "rule-evaluated",
            },
        }
    )


@dataclass
class FakeKernel:
    decisions: list[str] = field(default_factory=lambda: ["ALLOW"])
    execute_calls: list[dict[str, Any]] = field(default_factory=list)
    append_calls: list[dict[str, Any]] = field(default_factory=list)

    def execute(self, **kwargs: Any) -> DecisionTrace:
        self.execute_calls.append(kwargs)
        decision = self.decisions.pop(0) if self.decisions else "ALLOW"
        return _trace(decision)

    async def execute_async(self, **kwargs: Any) -> DecisionTrace:
        self.execute_calls.append(kwargs)
        decision = self.decisions.pop(0) if self.decisions else "ALLOW"
        return _trace(decision)

    def append_session_event(self, **kwargs: Any) -> dict[str, Any]:
        self.append_calls.append(kwargs)
        return {"ok": True}

    async def append_session_event_async(self, **kwargs: Any) -> dict[str, Any]:
        self.append_calls.append(kwargs)
        return {"ok": True}


def test_authorize_sets_skip_connector_metadata() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    trace = guard.authorize(
        provider="openai",
        operation="chat.completions",
        model="gpt-4o-mini",
        request={"messages": [{"role": "user", "content": "hello"}]},
    )

    assert trace.decision == "ALLOW"
    call = kernel.execute_calls[0]
    assert call["metadata"]["skip_connector"] is True
    assert call["intent"]["tool"] == "llm.openai.chat.completions"


def test_run_allow_executes_provider_and_records_event() -> None:
    kernel = FakeKernel(decisions=["ALLOW"])
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    ran = {"called": False}

    def invoke() -> dict[str, Any]:
        ran["called"] = True
        return {"id": "resp-1", "model": "gpt-4o-mini", "usage": {"total_tokens": 10}}

    result = guard.run(
        provider="openai",
        operation="chat.completions",
        model="gpt-4o-mini",
        request={"messages": [{"role": "user", "content": "hello"}]},
        invoke=invoke,
    )

    assert ran["called"] is True
    assert result.trace.decision == "ALLOW"
    assert result.response["id"] == "resp-1"
    assert kernel.append_calls and kernel.append_calls[0]["event_type"] == "llm.response"


def test_run_deny_raises() -> None:
    kernel = FakeKernel(decisions=["DENY"])
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    with pytest.raises(LLMPolicyDeniedError):
        guard.run(
            provider="openai",
            operation="chat.completions",
            model="gpt-4o-mini",
            request={"messages": [{"role": "user", "content": "hello"}]},
            invoke=lambda: {"id": "should-not-run"},
        )


def test_run_escalate_raises() -> None:
    kernel = FakeKernel(decisions=["ESCALATE"])
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    with pytest.raises(LLMPolicyEscalationError):
        guard.run(
            provider="openai",
            operation="chat.completions",
            model="gpt-4o-mini",
            request={"messages": [{"role": "user", "content": "hello"}]},
            invoke=lambda: {"id": "should-not-run"},
        )


def test_guard_openai_chat_completion_helper() -> None:
    kernel = FakeKernel(decisions=["ALLOW"])
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    def fake_create(*, model: str, messages: list[dict[str, Any]], temperature: float) -> dict[str, Any]:
        return {"id": "r-1", "model": model, "messages_count": len(messages), "temperature": temperature}

    result = guard_openai_chat_completion(
        guard=guard,
        create=fake_create,
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": "hello"}],
        temperature=0.3,
    )
    assert result.response["id"] == "r-1"
    assert kernel.execute_calls[0]["intent"]["tool"] == "llm.openai.chat.completions"


def test_run_async_allow() -> None:
    kernel = FakeKernel(decisions=["ALLOW"])
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    async def invoke_async() -> dict[str, Any]:
        return {"id": "async-1", "model": "claude-3-5"}

    result = asyncio.run(
        guard.run_async(
            provider="anthropic",
            operation="messages.create",
            model="claude-3-5-sonnet",
            request={"messages": [{"role": "user", "content": "hello"}]},
            invoke=invoke_async,
        )
    )
    assert result.response["id"] == "async-1"
    assert kernel.append_calls and kernel.append_calls[0]["event_type"] == "llm.response"


def test_authorized_call_context_manager() -> None:
    kernel = FakeKernel(decisions=["ALLOW"])
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    with guard.authorized_call(
        provider="openai",
        operation="chat.completions",
        model="gpt-4o-mini",
        request={"messages": [{"role": "user", "content": "hello"}]},
    ) as trace:
        response = {"id": "resp-ctx", "model": "gpt-4o-mini"}
        guard.record_response(provider="openai", operation="chat.completions", trace=trace, response=response)

    assert kernel.append_calls and kernel.append_calls[0]["payload"]["summary"]["id"] == "resp-ctx"



from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, ConfigDict

TraceArtifactKind = Literal["file", "connector", "domain", "output", "handoff", "other"]
TraceArtifactStatus = Literal["observed", "produced", "consumed", "blocked"]
TraceArtifactAction = Literal["read", "write", "call", "egress", "produce", "reuse", "handoff", "blocked"]


class TraceArtifactLineageSummary(BaseModel):
    model_config = ConfigDict(extra="allow")

    artifact_count: int
    side_effect_count: int
    blocked_count: int
    kinds: dict[str, int]
    top_domains: list[str] | None = None
    top_connectors: list[str] | None = None
    top_outputs: list[str] | None = None
    has_handoff: bool | None = None
    has_writes: bool | None = None
    has_blocked_side_effects: bool | None = None


class TraceArtifactRef(BaseModel):
    model_config = ConfigDict(extra="allow")

    artifact_id: str
    kind: TraceArtifactKind
    label: str
    locator: str | None = None
    status: TraceArtifactStatus | None = None
    chronos_entity_id: str | None = None
    chronos_anchor: str | None = None
    metadata: dict[str, Any] | None = None


class TraceArtifactLineageEvent(BaseModel):
    model_config = ConfigDict(extra="allow")

    event_id: str
    artifact_id: str
    action: TraceArtifactAction
    actor: str | None = None
    tool: str | None = None
    timestamp: str | None = None
    related_artifact_id: str | None = None
    metadata: dict[str, Any] | None = None


class TraceArtifactLineageResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    trace_id: str
    summary: TraceArtifactLineageSummary
    artifacts: list[TraceArtifactRef]
    events: list[TraceArtifactLineageEvent]



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import Any

import pytest

from sentinos.integrations import (
    AsyncOpenResponsesAdapter,
    LLMGuard,
    LLMPolicyDeniedError,
    OpenResponsesAdapter,
    create_async_openresponses_adapter,
    create_async_openrouter_openresponses_adapter,
    create_openresponses_adapter,
    create_openrouter_openresponses_adapter,
    guard_openresponses_create,
    iter_openresponses_sse_lines,
    parse_openresponses_sse_event,
)
from sentinos.models.decision_trace import DecisionTrace


def _trace(decision: str) -> DecisionTrace:
    return DecisionTrace.model_validate(
        {
            "trace_id": "trace-openresponses-1",
            "timestamp": "2026-02-07T00:00:00Z",
            "tenant_id": "acme",
            "agent_id": "agent-1",
            "session_id": "sess-1",
            "intent": {"type": "llm_call", "tool": "llm.openresponses.responses.create", "args": {}},
            "policy_evaluation": {
                "policy_id": "llm-guard",
                "policy_version": "v1",
                "decision": decision,
                "reason": "rule-evaluated",
            },
        }
    )


@dataclass
class FakeKernel:
    decision: str = "ALLOW"
    last_execute: dict[str, Any] | None = None
    events: list[dict[str, Any]] = field(default_factory=list)

    def execute(self, **kwargs: Any) -> DecisionTrace:
        self.last_execute = kwargs
        return _trace(self.decision)

    async def execute_async(self, **kwargs: Any) -> DecisionTrace:
        self.last_execute = kwargs
        return _trace(self.decision)

    def append_session_event(self, **kwargs: Any) -> dict[str, Any]:
        self.events.append(kwargs)
        return {"ok": True}

    async def append_session_event_async(self, **kwargs: Any) -> dict[str, Any]:
        self.events.append(kwargs)
        return {"ok": True}


def test_guard_openresponses_create_helper() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    result = guard_openresponses_create(
        guard=guard,
        create=lambda **kwargs: {"id": "resp-1", "status": "completed", **kwargs},
        model="gpt-4.1-mini",
        input=[{"type": "message", "role": "user", "content": "hello"}],
    )
    assert result.response.id == "resp-1"
    assert kernel.last_execute is not None
    assert kernel.last_execute["intent"]["tool"] == "llm.openresponses.responses.create"


def test_openresponses_adapter_create_and_request() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")
    adapter = OpenResponsesAdapter(
        guard=guard,
        create_fn=lambda **kwargs: {"id": "resp-2", "status": "completed", **kwargs},
    )

    result = adapter.create(
        model="gpt-4.1-mini",
        input=[{"type": "message", "role": "user", "content": "summarize"}],
        tools=[{"type": "function", "name": "get_status"}],
        tool_choice="auto",
    )
    assert result.response.id == "resp-2"
    assert result.response.model == "gpt-4.1-mini"

    result_2 = adapter.create_request(
        request={
            "model": "gpt-4.1-mini",
            "input": [{"type": "message", "role": "user", "content": "follow up"}],
            "metadata": {"tenant": "acme"},
        }
    )
    assert result_2.response.status == "completed"


def test_openresponses_stream_collects_events_and_final_response() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    lines = [
        "event: response.created",
        'data: {"type":"response.created","sequence_number":1}',
        "",
        "event: response.completed",
        (
            'data: {"type":"response.completed","sequence_number":2,'
            '"response":{"id":"resp-stream","status":"completed","model":"gpt-4.1-mini","output":[]}}'
        ),
        "",
        "data: [DONE]",
        "",
    ]

    adapter = OpenResponsesAdapter(
        guard=guard,
        create_fn=lambda **kwargs: {"id": "unused", **kwargs},
        stream_fn=lambda **kwargs: lines,
    )
    result = adapter.stream(
        model="gpt-4.1-mini",
        input=[{"type": "message", "role": "user", "content": "hello"}],
    )
    assert result.decision == "ALLOW"
    assert len(result.events) >= 2
    assert result.final_response is not None
    assert result.final_response.id == "resp-stream"
    assert kernel.events and kernel.events[0]["event_type"] == "llm.response"


def test_openresponses_stream_denied_raises() -> None:
    kernel = FakeKernel(decision="DENY")
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    adapter = OpenResponsesAdapter(
        guard=guard,
        create_fn=lambda **kwargs: {"id": "unused", **kwargs},
        stream_fn=lambda **kwargs: [],
    )
    with pytest.raises(LLMPolicyDeniedError):
        adapter.stream(model="gpt-4.1-mini", input=[{"type": "message", "role": "user", "content": "hello"}])


def test_sse_helpers_support_json_lines_and_done() -> None:
    events = list(
        iter_openresponses_sse_lines(
            [
                '{"type":"response.created","sequence_number":1}',
                "",
                "data: [DONE]",
                "",
            ]
        )
    )
    assert events[0].type == "response.created"
    assert events[-1].type == "response.done"

    event = parse_openresponses_sse_event('{"type":"response.failed","sequence_number":9}')
    assert event is not None
    assert event.type == "response.failed"


def test_openresponses_factory_with_explicit_clients() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    class _Responses:
        @staticmethod
        def create(**kwargs: Any) -> dict[str, Any]:
            return {"id": "factory-sync", "status": "completed", **kwargs}

        @staticmethod
        def stream(**kwargs: Any) -> list[dict[str, Any]]:
            return [
                {"type": "response.created", "sequence_number": 1},
                {
                    "type": "response.completed",
                    "sequence_number": 2,
                    "response": {"id": "factory-stream", "status": "completed", **kwargs},
                },
            ]

    class _Client:
        responses = _Responses()

    adapter = create_openresponses_adapter(guard=guard, client=_Client())
    out = adapter.create(model="gpt-4.1-mini", input="hello")
    assert out.response.id == "factory-sync"
    stream_out = adapter.stream(model="gpt-4.1-mini", input="hello")
    assert stream_out.final_response is not None
    assert stream_out.final_response.id == "factory-stream"


def test_async_openresponses_adapter() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    async def create_fn(**kwargs: Any) -> dict[str, Any]:
        return {"id": "async-created", "status": "completed", **kwargs}

    async def stream_fn(**kwargs: Any) -> list[dict[str, Any]]:
        return [
            {"type": "response.created", "sequence_number": 1},
            {
                "type": "response.completed",
                "sequence_number": 2,
                "response": {"id": "async-stream", "status": "completed", **kwargs},
            },
        ]

    adapter = AsyncOpenResponsesAdapter(guard=guard, create_fn=create_fn, stream_fn=stream_fn)
    created = asyncio.run(adapter.create(model="gpt-4.1-mini", input="hello"))
    assert created.response.id == "async-created"

    streamed = asyncio.run(adapter.stream(model="gpt-4.1-mini", input="hello"))
    assert streamed.final_response is not None
    assert streamed.final_response.id == "async-stream"


def test_async_factory_with_explicit_client() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    class _Responses:
        @staticmethod
        async def create(**kwargs: Any) -> dict[str, Any]:
            return {"id": "async-factory", "status": "completed", **kwargs}

    class _Client:
        responses = _Responses()

    adapter = create_async_openresponses_adapter(guard=guard, client=_Client())
    out = asyncio.run(adapter.create(model="gpt-4.1-mini", input="hello"))
    assert out.response.id == "async-factory"


def test_openresponses_factory_provider_override() -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    class _Responses:
        @staticmethod
        def create(**kwargs: Any) -> dict[str, Any]:
            return {"id": "override-sync", "status": "completed", **kwargs}

    class _Client:
        responses = _Responses()

    adapter = create_openresponses_adapter(
        guard=guard,
        client=_Client(),
        provider="openrouter",
        operation="responses.create",
    )
    out = adapter.create(model="openai/gpt-4.1-mini", input="hello")
    assert out.response.id == "override-sync"
    assert kernel.last_execute is not None
    assert kernel.last_execute["intent"]["tool"] == "llm.openrouter.responses.create"


def test_openrouter_openresponses_factory_defaults_and_headers(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = FakeKernel()
    guard = LLMGuard(kernel=kernel, agent_id="agent-1", session_id="sess-1", tenant_id="acme")

    created_kwargs: dict[str, Any] = {}

    class _Responses:
        @staticmethod
        def create(**kwargs: Any) -> dict[str, Any]:
            return {"id": "or-sync", "status": "completed", **kwargs}

    class _AsyncResponses:
        @staticmethod
        async def create(**kwargs: Any) -> dict[str, Any]:
            return {"id": "or-async", "status": "completed", **kwargs}

    class _OpenAIClient:
        responses = _Responses()

        def __init__(self, **kwargs: Any):
            created_kwargs.update(kwargs)

    class _AsyncOpenAIClient:
        responses = _AsyncResponses()

        def __init__(self, **kwargs: Any):
            created_kwargs.update(kwargs)

    class _OpenAIModule:
        OpenAI = _OpenAIClient
        AsyncOpenAI = _AsyncOpenAIClient

    monkeypatch.setattr(
        "sentinos.integrations.openresponses.importlib.import_module",
        lambda name: _OpenAIModule(),
    )
    monkeypatch.setenv("OPENROUTER_API_KEY", "openrouter-env-key")

    adapter = create_openrouter_openresponses_adapter(
        guard=guard,
        http_referer="https://console.sentinos.dev",
        x_title="Sentinos Console",
    )
    out = adapter.create(model="openai/gpt-4.1-mini", input="hello")
    assert out.response.id == "or-sync"
    assert kernel.last_execute is not None
    assert kernel.last_execute["intent"]["tool"] == "llm.openrouter.responses.create"
    assert created_kwargs["base_url"] == "https://openrouter.ai/api/v1"
    assert created_kwargs["api_key"] == "openrouter-env-key"
    assert created_kwargs["default_headers"]["HTTP-Referer"] == "https://console.sentinos.dev"
    assert created_kwargs["default_headers"]["X-Title"] == "Sentinos Console"

    async_adapter = create_async_openrouter_openresponses_adapter(guard=guard)
    out_async = asyncio.run(async_adapter.create(model="openai/gpt-4.1-mini", input="hello"))
    assert out_async.response.id == "or-async"
    assert kernel.last_execute is not None
    assert kernel.last_execute["intent"]["tool"] == "llm.openrouter.responses.create"



from __future__ import annotations

import asyncio
import contextlib
import importlib
from dataclasses import dataclass
from typing import Any

from ..models.decision_trace import DecisionTrace
from .llm import LLMGuard, LLMPolicyDeniedError, LLMPolicyEscalationError, LLMPolicyResult


class BedrockProviderError(RuntimeError):
    """Base provider-side error raised when Bedrock invocation fails after policy authorization."""

    def __init__(self, message: str, *, error_code: str | None = None, raw: Any | None = None):
        super().__init__(message)
        self.error_code = error_code
        self.raw = raw


class BedrockAccessDeniedError(BedrockProviderError):
    """Raised when Bedrock denies access to the target model or API."""


class BedrockThrottlingError(BedrockProviderError):
    """Raised when Bedrock throttles the request."""


class BedrockValidationError(BedrockProviderError):
    """Raised when Bedrock rejects request payload shape or unsupported fields."""


@dataclass(frozen=True)
class BedrockConverseStreamResult:
    provider: str
    operation: str
    decision: str
    trace: DecisionTrace
    events: list[dict[str, Any]]
    output_text: str
    stop_reason: str | None
    usage: dict[str, Any] | None
    metrics: dict[str, Any] | None


def _import_optional(module_name: str, extra_name: str) -> Any:
    try:
        return importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        raise ImportError(
            f"Optional dependency '{module_name}' is not installed. Install with: pip install 'sentinos[{extra_name}]'"
        ) from exc


def _decision_text(trace: DecisionTrace) -> str:
    return str(trace.decision).strip().upper()


def _ensure_allowed(*, provider: str, operation: str, trace: DecisionTrace) -> str:
    decision = _decision_text(trace)
    reason = trace.policy_evaluation.reason or "policy decision"
    if decision == "DENY":
        raise LLMPolicyDeniedError(f"Sentinos denied {provider}.{operation}: {reason}", trace=trace)
    if decision == "ESCALATE":
        raise LLMPolicyEscalationError(f"Sentinos escalated {provider}.{operation}: {reason}", trace=trace)
    return decision


def _to_payload(value: Any) -> dict[str, Any]:
    if isinstance(value, dict):
        return value
    if hasattr(value, "model_dump") and callable(getattr(value, "model_dump")):
        return getattr(value, "model_dump")()
    if hasattr(value, "to_dict") and callable(getattr(value, "to_dict")):
        return getattr(value, "to_dict")()
    if hasattr(value, "dict") and callable(getattr(value, "dict")):
        return getattr(value, "dict")()
    if hasattr(value, "__dict__"):
        return {str(k): v for k, v in vars(value).items() if not str(k).startswith("_")}
    raise TypeError(f"Unable to normalize payload from type {type(value).__name__}")


def _extract_error_code(exc: Exception) -> tuple[str | None, str]:
    message = str(exc)
    response = getattr(exc, "response", None)
    if isinstance(response, dict):
        err = response.get("Error")
        if isinstance(err, dict):
            code = err.get("Code")
            if isinstance(err.get("Message"), str) and err["Message"]:
                message = err["Message"]
            if isinstance(code, str) and code:
                return code, message
    code_attr = getattr(exc, "code", None)
    if isinstance(code_attr, str) and code_attr:
        return code_attr, message
    return None, message


def _raise_mapped_bedrock_error(exc: Exception) -> None:
    code, message = _extract_error_code(exc)
    prefix = f"Bedrock error {code}" if code else "Bedrock error"
    full = f"{prefix}: {message}"

    if code == "AccessDeniedException":
        raise BedrockAccessDeniedError(full, error_code=code, raw=exc) from exc
    if code == "ThrottlingException":
        raise BedrockThrottlingError(full, error_code=code, raw=exc) from exc
    if code == "ValidationException":
        raise BedrockValidationError(full, error_code=code, raw=exc) from exc
    raise BedrockProviderError(full, error_code=code, raw=exc) from exc


def _build_converse_payload(
    *,
    model_id: str,
    messages: list[dict[str, Any]],
    system: list[dict[str, Any]] | None = None,
    inference_config: dict[str, Any] | None = None,
    tool_config: dict[str, Any] | None = None,
    guardrail_config: dict[str, Any] | None = None,
    additional_model_request_fields: dict[str, Any] | None = None,
    additional_model_response_field_paths: list[str] | None = None,
    request_metadata: dict[str, str] | None = None,
    performance_config: dict[str, Any] | None = None,
    prompt_variables: dict[str, Any] | None = None,
    extra_fields: dict[str, Any] | None = None,
) -> dict[str, Any]:
    payload: dict[str, Any] = {
        "modelId": model_id,
        "messages": messages,
    }
    if system is not None:
        payload["system"] = system
    if inference_config is not None:
        payload["inferenceConfig"] = inference_config
    if tool_config is not None:
        payload["toolConfig"] = tool_config
    if guardrail_config is not None:
        payload["guardrailConfig"] = guardrail_config
    if additional_model_request_fields is not None:
        payload["additionalModelRequestFields"] = additional_model_request_fields
    if additional_model_response_field_paths is not None:
        payload["additionalModelResponseFieldPaths"] = additional_model_response_field_paths
    if request_metadata is not None:
        payload["requestMetadata"] = request_metadata
    if performance_config is not None:
        payload["performanceConfig"] = performance_config
    if prompt_variables is not None:
        payload["promptVariables"] = prompt_variables
    if extra_fields:
        payload.update(extra_fields)
    return payload


def _summarize_converse_response(response: dict[str, Any]) -> dict[str, Any]:
    output_message = response.get("output", {}).get("message") if isinstance(response.get("output"), dict) else None
    output_text: list[str] = []
    if isinstance(output_message, dict):
        content = output_message.get("content")
        if isinstance(content, list):
            for item in content:
                if isinstance(item, dict):
                    text = item.get("text")
                    if isinstance(text, str):
                        output_text.append(text)

    summary: dict[str, Any] = {
        "stopReason": response.get("stopReason"),
        "output_text": "".join(output_text),
    }
    usage = response.get("usage")
    if isinstance(usage, dict):
        summary["usage"] = usage
    metrics = response.get("metrics")
    if isinstance(metrics, dict):
        summary["metrics"] = metrics
    return summary


def _collect_stream(
    raw_response: Any,
) -> tuple[list[dict[str, Any]], str, str | None, dict[str, Any] | None, dict[str, Any] | None]:
    response_dict = _to_payload(raw_response)
    raw_stream = response_dict.get("stream")
    if raw_stream is None:
        return [], "", None, None, None

    events: list[dict[str, Any]] = []
    output_parts: list[str] = []
    stop_reason: str | None = None
    usage: dict[str, Any] | None = None
    metrics: dict[str, Any] | None = None

    for raw_event in raw_stream:
        event = _to_payload(raw_event)
        events.append(event)

        content_delta = event.get("contentBlockDelta")
        if isinstance(content_delta, dict):
            delta = content_delta.get("delta")
            if isinstance(delta, dict):
                text = delta.get("text")
                if isinstance(text, str):
                    output_parts.append(text)

        message_stop = event.get("messageStop")
        if isinstance(message_stop, dict):
            reason = message_stop.get("stopReason")
            if isinstance(reason, str):
                stop_reason = reason

        metadata = event.get("metadata")
        if isinstance(metadata, dict):
            maybe_usage = metadata.get("usage")
            if isinstance(maybe_usage, dict):
                usage = maybe_usage
            maybe_metrics = metadata.get("metrics")
            if isinstance(maybe_metrics, dict):
                metrics = maybe_metrics

    output_text = "".join(output_parts)
    return events, output_text, stop_reason, usage, metrics


@dataclass
class BedrockConverseAdapter:
    guard: LLMGuard
    client: Any
    provider: str = "bedrock"

    def converse(
        self,
        *,
        model_id: str,
        messages: list[dict[str, Any]],
        system: list[dict[str, Any]] | None = None,
        inference_config: dict[str, Any] | None = None,
        tool_config: dict[str, Any] | None = None,
        guardrail_config: dict[str, Any] | None = None,
        additional_model_request_fields: dict[str, Any] | None = None,
        additional_model_response_field_paths: list[str] | None = None,
        request_metadata: dict[str, str] | None = None,
        performance_config: dict[str, Any] | None = None,
        prompt_variables: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> LLMPolicyResult[dict[str, Any]]:
        payload = _build_converse_payload(
            model_id=model_id,
            messages=messages,
            system=system,
            inference_config=inference_config,
            tool_config=tool_config,
            guardrail_config=guardrail_config,
            additional_model_request_fields=additional_model_request_fields,
            additional_model_response_field_paths=additional_model_response_field_paths,
            request_metadata=request_metadata,
            performance_config=performance_config,
            prompt_variables=prompt_variables,
            extra_fields=kwargs,
        )
        trace = self.guard.authorize(
            provider=self.provider,
            operation="converse",
            request=payload,
            model=model_id,
            metadata=metadata,
            tenant_id=tenant_id,
        )
        _ensure_allowed(provider=self.provider, operation="converse", trace=trace)

        try:
            raw_response = self.client.converse(**payload)
        except Exception as exc:
            _raise_mapped_bedrock_error(exc)

        response = _to_payload(raw_response)
        with contextlib.suppress(Exception):
            self.guard.record_response(
                provider=self.provider,
                operation="converse",
                trace=trace,
                response=response,
                tenant_id=tenant_id,
                response_summarizer=_summarize_converse_response,
            )
        return LLMPolicyResult(provider=self.provider, operation="converse", trace=trace, response=response)

    def converse_stream(
        self,
        *,
        model_id: str,
        messages: list[dict[str, Any]],
        system: list[dict[str, Any]] | None = None,
        inference_config: dict[str, Any] | None = None,
        tool_config: dict[str, Any] | None = None,
        guardrail_config: dict[str, Any] | None = None,
        additional_model_request_fields: dict[str, Any] | None = None,
        additional_model_response_field_paths: list[str] | None = None,
        request_metadata: dict[str, str] | None = None,
        performance_config: dict[str, Any] | None = None,
        prompt_variables: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> BedrockConverseStreamResult:
        payload = _build_converse_payload(
            model_id=model_id,
            messages=messages,
            system=system,
            inference_config=inference_config,
            tool_config=tool_config,
            guardrail_config=guardrail_config,
            additional_model_request_fields=additional_model_request_fields,
            additional_model_response_field_paths=additional_model_response_field_paths,
            request_metadata=request_metadata,
            performance_config=performance_config,
            prompt_variables=prompt_variables,
            extra_fields=kwargs,
        )
        trace = self.guard.authorize(
            provider=self.provider,
            operation="converse_stream",
            request=payload,
            model=model_id,
            metadata=metadata,
            tenant_id=tenant_id,
        )
        decision = _ensure_allowed(provider=self.provider, operation="converse_stream", trace=trace)

        try:
            raw_response = self.client.converse_stream(**payload)
            events, output_text, stop_reason, usage, metrics = _collect_stream(raw_response)
        except Exception as exc:
            _raise_mapped_bedrock_error(exc)

        with contextlib.suppress(Exception):
            self.guard.record_response(
                provider=self.provider,
                operation="converse_stream",
                trace=trace,
                response={
                    "output_text": output_text,
                    "stopReason": stop_reason,
                    "usage": usage,
                    "metrics": metrics,
                    "events": len(events),
                },
                tenant_id=tenant_id,
            )
        return BedrockConverseStreamResult(
            provider=self.provider,
            operation="converse_stream",
            decision=decision,
            trace=trace,
            events=events,
            output_text=output_text,
            stop_reason=stop_reason,
            usage=usage,
            metrics=metrics,
        )


@dataclass
class AsyncBedrockConverseAdapter:
    guard: LLMGuard
    client: Any
    provider: str = "bedrock"

    async def converse(
        self,
        *,
        model_id: str,
        messages: list[dict[str, Any]],
        system: list[dict[str, Any]] | None = None,
        inference_config: dict[str, Any] | None = None,
        tool_config: dict[str, Any] | None = None,
        guardrail_config: dict[str, Any] | None = None,
        additional_model_request_fields: dict[str, Any] | None = None,
        additional_model_response_field_paths: list[str] | None = None,
        request_metadata: dict[str, str] | None = None,
        performance_config: dict[str, Any] | None = None,
        prompt_variables: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> LLMPolicyResult[dict[str, Any]]:
        payload = _build_converse_payload(
            model_id=model_id,
            messages=messages,
            system=system,
            inference_config=inference_config,
            tool_config=tool_config,
            guardrail_config=guardrail_config,
            additional_model_request_fields=additional_model_request_fields,
            additional_model_response_field_paths=additional_model_response_field_paths,
            request_metadata=request_metadata,
            performance_config=performance_config,
            prompt_variables=prompt_variables,
            extra_fields=kwargs,
        )
        trace = await self.guard.authorize_async(
            provider=self.provider,
            operation="converse",
            request=payload,
            model=model_id,
            metadata=metadata,
            tenant_id=tenant_id,
        )
        _ensure_allowed(provider=self.provider, operation="converse", trace=trace)

        try:
            raw_response = await asyncio.to_thread(self.client.converse, **payload)
        except Exception as exc:
            _raise_mapped_bedrock_error(exc)

        response = _to_payload(raw_response)
        with contextlib.suppress(Exception):
            await self.guard.record_response_async(
                provider=self.provider,
                operation="converse",
                trace=trace,
                response=response,
                tenant_id=tenant_id,
                response_summarizer=_summarize_converse_response,
            )
        return LLMPolicyResult(provider=self.provider, operation="converse", trace=trace, response=response)

    async def converse_stream(
        self,
        *,
        model_id: str,
        messages: list[dict[str, Any]],
        system: list[dict[str, Any]] | None = None,
        inference_config: dict[str, Any] | None = None,
        tool_config: dict[str, Any] | None = None,
        guardrail_config: dict[str, Any] | None = None,
        additional_model_request_fields: dict[str, Any] | None = None,
        additional_model_response_field_paths: list[str] | None = None,
        request_metadata: dict[str, str] | None = None,
        performance_config: dict[str, Any] | None = None,
        prompt_variables: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> BedrockConverseStreamResult:
        payload = _build_converse_payload(
            model_id=model_id,
            messages=messages,
            system=system,
            inference_config=inference_config,
            tool_config=tool_config,
            guardrail_config=guardrail_config,
            additional_model_request_fields=additional_model_request_fields,
            additional_model_response_field_paths=additional_model_response_field_paths,
            request_metadata=request_metadata,
            performance_config=performance_config,
            prompt_variables=prompt_variables,
            extra_fields=kwargs,
        )
        trace = await self.guard.authorize_async(
            provider=self.provider,
            operation="converse_stream",
            request=payload,
            model=model_id,
            metadata=metadata,
            tenant_id=tenant_id,
        )
        decision = _ensure_allowed(provider=self.provider, operation="converse_stream", trace=trace)

        try:
            raw_response = await asyncio.to_thread(self.client.converse_stream, **payload)
            events, output_text, stop_reason, usage, metrics = await asyncio.to_thread(_collect_stream, raw_response)
        except Exception as exc:
            _raise_mapped_bedrock_error(exc)

        with contextlib.suppress(Exception):
            await self.guard.record_response_async(
                provider=self.provider,
                operation="converse_stream",
                trace=trace,
                response={
                    "output_text": output_text,
                    "stopReason": stop_reason,
                    "usage": usage,
                    "metrics": metrics,
                    "events": len(events),
                },
                tenant_id=tenant_id,
            )
        return BedrockConverseStreamResult(
            provider=self.provider,
            operation="converse_stream",
            decision=decision,
            trace=trace,
            events=events,
            output_text=output_text,
            stop_reason=stop_reason,
            usage=usage,
            metrics=metrics,
        )


def _create_bedrock_runtime_client(
    *,
    client: Any | None = None,
    session: Any | None = None,
    region_name: str | None = None,
    profile_name: str | None = None,
    endpoint_url: str | None = None,
    **client_kwargs: Any,
) -> Any:
    if client is not None:
        return client

    boto3_module = _import_optional("boto3", "bedrock")

    resolved_session = session
    if resolved_session is None:
        session_kwargs: dict[str, Any] = {}
        if profile_name is not None:
            session_kwargs["profile_name"] = profile_name
        if region_name is not None:
            session_kwargs["region_name"] = region_name
        resolved_session = boto3_module.Session(**session_kwargs)

    kwargs: dict[str, Any] = dict(client_kwargs)
    if endpoint_url is not None:
        kwargs["endpoint_url"] = endpoint_url
    if region_name is not None and session is not None:
        kwargs["region_name"] = region_name

    return resolved_session.client("bedrock-runtime", **kwargs)


def create_bedrock_converse_adapter(
    *,
    guard: LLMGuard,
    client: Any | None = None,
    session: Any | None = None,
    region_name: str | None = None,
    profile_name: str | None = None,
    endpoint_url: str | None = None,
    provider: str = "bedrock",
    **client_kwargs: Any,
) -> BedrockConverseAdapter:
    runtime_client = _create_bedrock_runtime_client(
        client=client,
        session=session,
        region_name=region_name,
        profile_name=profile_name,
        endpoint_url=endpoint_url,
        **client_kwargs,
    )
    return BedrockConverseAdapter(guard=guard, client=runtime_client, provider=provider)


def create_async_bedrock_converse_adapter(
    *,
    guard: LLMGuard,
    client: Any | None = None,
    session: Any | None = None,
    region_name: str | None = None,
    profile_name: str | None = None,
    endpoint_url: str | None = None,
    provider: str = "bedrock",
    **client_kwargs: Any,
) -> AsyncBedrockConverseAdapter:
    runtime_client = _create_bedrock_runtime_client(
        client=client,
        session=session,
        region_name=region_name,
        profile_name=profile_name,
        endpoint_url=endpoint_url,
        **client_kwargs,
    )
    return AsyncBedrockConverseAdapter(guard=guard, client=runtime_client, provider=provider)



from __future__ import annotations

from sentinos.client import SentinosClient


def test_sentinos_client_base_url_fans_out_to_all_services() -> None:
    client = SentinosClient(base_url="https://api.sentinos.ai")

    assert client.config.base_url == "https://api.sentinos.ai"
    assert client.config.kernel_url == "https://api.sentinos.ai"
    assert client.config.arbiter_url == "https://api.sentinos.ai"
    assert client.config.chronos_url == "https://api.sentinos.ai"
    assert client.config.controlplane_url == "https://api.sentinos.ai"
    assert client.config.meshgate_url == "https://api.sentinos.ai"
    assert client.controlplane is not None
    assert client.audit is not None
    assert client.dashboards is not None
    assert client.billing is not None
    assert client.privacy is not None
    assert client.a2a is not None


def test_sentinos_client_service_url_overrides_base_url() -> None:
    client = SentinosClient(
        base_url="https://api.sentinos.ai",
        kernel_url="https://kernel.sentinos.ai",
        arbiter_url="https://arbiter.sentinos.ai/",
        controlplane_url="https://controlplane.sentinos.ai/",
    )

    assert client.config.base_url == "https://api.sentinos.ai"
    assert client.config.kernel_url == "https://kernel.sentinos.ai"
    assert client.config.arbiter_url == "https://arbiter.sentinos.ai"
    assert client.config.chronos_url == "https://api.sentinos.ai"
    assert client.config.controlplane_url == "https://controlplane.sentinos.ai"
    assert client.config.meshgate_url == "https://api.sentinos.ai"


def test_sentinos_client_default_localhost_urls_when_base_not_set() -> None:
    client = SentinosClient()

    assert client.config.base_url is None
    assert client.config.kernel_url == "http://localhost:8081"
    assert client.config.arbiter_url == "http://localhost:8082"
    assert client.config.chronos_url == "http://localhost:8083"
    assert client.config.controlplane_url == "http://localhost:18084"
    assert client.config.meshgate_url == "http://localhost:8085"


def test_sentinos_client_local_base_url_does_not_override_controlplane_default() -> None:
    client = SentinosClient(base_url="http://localhost:8081")

    assert client.config.kernel_url == "http://localhost:8081"
    assert client.config.arbiter_url == "http://localhost:8081"
    assert client.config.chronos_url == "http://localhost:8081"
    assert client.config.controlplane_url == "http://localhost:18084"
    assert client.config.meshgate_url == "http://localhost:8085"


def test_sentinos_client_api_url_alias_fans_out() -> None:
    client = SentinosClient(api_url="https://api.sentinos.ai/")

    assert client.config.base_url == "https://api.sentinos.ai"
    assert client.config.kernel_url == "https://api.sentinos.ai"
    assert client.config.arbiter_url == "https://api.sentinos.ai"
    assert client.config.chronos_url == "https://api.sentinos.ai"
    assert client.config.controlplane_url == "https://api.sentinos.ai"
    assert client.config.meshgate_url == "https://api.sentinos.ai"


def test_sentinos_client_url_alias_fans_out() -> None:
    client = SentinosClient(url="https://api.sentinos.ai/")

    assert client.config.base_url == "https://api.sentinos.ai"
    assert client.config.kernel_url == "https://api.sentinos.ai"
    assert client.config.arbiter_url == "https://api.sentinos.ai"
    assert client.config.chronos_url == "https://api.sentinos.ai"
    assert client.config.controlplane_url == "https://api.sentinos.ai"
    assert client.config.meshgate_url == "https://api.sentinos.ai"


def test_sentinos_client_base_url_precedence_over_aliases() -> None:
    client = SentinosClient(
        url="https://url.sentinos.ai",
        api_url="https://api.sentinos.ai",
        base_url="https://base.sentinos.ai",
    )

    assert client.config.base_url == "https://base.sentinos.ai"
    assert client.config.kernel_url == "https://base.sentinos.ai"
    assert client.config.arbiter_url == "https://base.sentinos.ai"
    assert client.config.chronos_url == "https://base.sentinos.ai"
    assert client.config.controlplane_url == "https://base.sentinos.ai"
    assert client.config.meshgate_url == "https://base.sentinos.ai"


def test_sentinos_client_simple_helper() -> None:
    client = SentinosClient.simple(
        base_url="https://api.sentinos.ai/",
        tenant_id="org-1",
        auth_token="token-1",
    )

    assert client.config.base_url == "https://api.sentinos.ai"
    assert client.config.tenant_id == "org-1"
    assert client.config.kernel_url == "https://api.sentinos.ai"
    assert client.config.controlplane_url == "https://api.sentinos.ai"
    assert client.config.meshgate_url == "https://api.sentinos.ai"


def test_sentinos_client_org_id_alias_sets_tenant_id() -> None:
    client = SentinosClient(base_url="https://api.sentinos.ai", org_id="org-2")
    assert client.config.tenant_id == "org-2"


def test_sentinos_client_org_id_alias_conflict_raises() -> None:
    try:
        SentinosClient(base_url="https://api.sentinos.ai", tenant_id="org-a", org_id="org-b")
    except ValueError as e:
        assert "tenant_id and org_id must match" in str(e)
    else:
        raise AssertionError("expected ValueError on tenant_id/org_id mismatch")


def test_sentinos_client_from_env_base_url(monkeypatch) -> None:
    monkeypatch.setenv("SENTINOS_BASE_URL", "https://api.sentinos.ai/")
    monkeypatch.setenv("SENTINOS_TENANT_ID", "tenant-a")
    monkeypatch.setenv("SENTINOS_ACCESS_TOKEN", "token-a")

    client = SentinosClient.from_env()

    assert client.config.base_url == "https://api.sentinos.ai"
    assert client.config.tenant_id == "tenant-a"
    assert client.config.kernel_url == "https://api.sentinos.ai"
    assert client.config.arbiter_url == "https://api.sentinos.ai"
    assert client.config.chronos_url == "https://api.sentinos.ai"
    assert client.config.controlplane_url == "https://api.sentinos.ai"
    assert client.config.meshgate_url == "https://api.sentinos.ai"


def test_sentinos_client_from_env_respects_service_overrides(monkeypatch) -> None:
    monkeypatch.setenv("SENTINOS_API_URL", "https://api.sentinos.ai")
    monkeypatch.setenv("SENTINOS_KERNEL_URL", "https://kernel.internal")
    monkeypatch.setenv("SENTINOS_ARBITER_URL", "https://arbiter.internal/")
    monkeypatch.setenv("SENTINOS_CHRONOS_URL", "https://chronos.internal")
    monkeypatch.setenv("SENTINOS_CONTROLPLANE_URL", "https://controlplane.internal/")
    monkeypatch.setenv("SENTINOS_MESHGATE_URL", "https://meshgate.internal/")
    monkeypatch.setenv("SENTINOS_ORG_ID", "org-via-alias")
    monkeypatch.setenv("SENTINOS_TIMEOUT_SECONDS", "45")

    client = SentinosClient.from_env()

    assert client.config.base_url == "https://api.sentinos.ai"
    assert client.config.kernel_url == "https://kernel.internal"
    assert client.config.arbiter_url == "https://arbiter.internal"
    assert client.config.chronos_url == "https://chronos.internal"
    assert client.config.controlplane_url == "https://controlplane.internal"
    assert client.config.meshgate_url == "https://meshgate.internal"
    assert client.config.tenant_id == "org-via-alias"
    assert client.config.timeout_seconds == 45.0


def test_sentinos_client_from_env_org_id_param_alias_sets_tenant_id(monkeypatch) -> None:
    monkeypatch.delenv("SENTINOS_TENANT_ID", raising=False)
    monkeypatch.delenv("SENTINOS_ORG_ID", raising=False)

    client = SentinosClient.from_env(org_id="org-param")
    assert client.config.tenant_id == "org-param"


def test_sentinos_client_from_env_org_id_conflict_raises() -> None:
    try:
        SentinosClient.from_env(tenant_id="org-a", org_id="org-b")
    except ValueError as e:
        assert "tenant_id and org_id must match" in str(e)
    else:
        raise AssertionError("expected ValueError on tenant_id/org_id mismatch")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from sentinos_core import AuthenticatedClient, Client

from ._http import request_json, request_json_async, require_tenant


@dataclass
class PrivacyClient:
    _core: Client | AuthenticatedClient
    tenant_id: str | None = None

    def _tenant(self, tenant_id: str | None) -> str:
        return require_tenant(self.tenant_id, tenant_id)

    def get_policy(self, *, tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(self._core, "GET", "/v1/trace/privacy/policy", tenant_id=self._tenant(tenant_id))

    def update_policy(self, *, policy: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "PATCH", "/v1/trace/privacy/policy", tenant_id=self._tenant(tenant_id), body=policy
        )

    async def get_policy_async(self, *, tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core, "GET", "/v1/trace/privacy/policy", tenant_id=self._tenant(tenant_id)
        )

    async def update_policy_async(self, *, policy: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "PATCH",
            "/v1/trace/privacy/policy",
            tenant_id=self._tenant(tenant_id),
            body=policy,
        )

    def scan(self, *, payload: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "POST", "/v1/trace/privacy/scan", tenant_id=self._tenant(tenant_id), body=payload
        )

    async def scan_async(self, *, payload: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core, "POST", "/v1/trace/privacy/scan", tenant_id=self._tenant(tenant_id), body=payload
        )



"""Provider-agnostic LLM governance using Sentinos LLMGuard."""

from __future__ import annotations

from sentinos import LLMGuard, SentinosClient


def fake_provider_call(*, model: str, messages: list[dict[str, str]], temperature: float) -> dict[str, object]:
    return {
        "id": "resp-123",
        "model": model,
        "usage": {"prompt_tokens": len(str(messages)), "completion_tokens": 42},
        "temperature": temperature,
    }


def run_any_provider_example() -> None:
    client = SentinosClient.from_env(org_id="acme")
    guard = LLMGuard(kernel=client.kernel, agent_id="support-agent", session_id="sess-001")

    result = guard.run(
        provider="openai",
        operation="chat.completions",
        model="gpt-4o-mini",
        request={
            "messages": [
                {"role": "system", "content": "You are concise and safe."},
                {"role": "user", "content": "Summarize yesterday's incident."},
            ]
        },
        invoke=lambda: fake_provider_call(
            model="gpt-4o-mini",
            messages=[{"role": "user", "content": "Summarize yesterday's incident."}],
            temperature=0.2,
        ),
    )
    print(result.trace.trace_id, result.trace.decision, result.response["id"])


if __name__ == "__main__":
    run_any_provider_example()



"""Quickstart SDK example."""

from sentinos import SentinosClient
from sentinos.auth.jwt import JWTAuth


def build_client(token: str) -> SentinosClient:
    # Configure SENTINOS_BASE_URL / SENTINOS_*_URL in your environment before running.
    return SentinosClient.from_env(org_id="acme", auth=JWTAuth(lambda: token))


def run_example() -> None:
    client = build_client("dev-token")
    traces = client.traces.list_traces()
    print(f"Connected. Traces returned: {len(traces)}")


if __name__ == "__main__":
    run_example()



from __future__ import annotations

from pathlib import Path

from sentinos.tools.sdk_migration_lint import lint_paths


def test_sdk_migration_lint_flags_tenant_id_first_constructor(tmp_path: Path) -> None:
    p = tmp_path / "example.py"
    p.write_text('from sentinos import SentinosClient\nclient = SentinosClient(tenant_id="acme")\n', encoding="utf-8")

    findings = lint_paths([str(tmp_path)])
    assert any(f.rule_id == "DX008-PY001" for f in findings)


def test_sdk_migration_lint_ignores_alias_documentation_lines(tmp_path: Path) -> None:
    p = tmp_path / "doc.md"
    p.write_text("export SENTINOS_ORG_ID=acme  # alias: SENTINOS_TENANT_ID\n", encoding="utf-8")

    findings = lint_paths([str(tmp_path)])
    assert not any(f.rule_id == "DX008-ENV001" for f in findings)


def test_sdk_migration_lint_flags_api_url_alias_usage(tmp_path: Path) -> None:
    p = tmp_path / "example.md"
    p.write_text('```python\nclient = SentinosClient(api_url="https://api.sentinos.ai")\n```\n', encoding="utf-8")

    findings = lint_paths([str(tmp_path)])
    assert any(f.rule_id == "DX008-PY002" for f in findings)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from sentinos_core import AuthenticatedClient, Client

from ._http import request_json, request_json_async


@dataclass
class A2AClient:
    _core: Client | AuthenticatedClient
    tenant_id: str | None = None

    def authorize_handoff(self, *, handoff: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "POST", "/v1/a2a/handoffs/authorize", tenant_id=tenant_id or self.tenant_id, body=handoff
        )

    async def authorize_handoff_async(self, *, handoff: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "POST",
            "/v1/a2a/handoffs/authorize",
            tenant_id=tenant_id or self.tenant_id,
            body=handoff,
        )

    def forward_handoff(self, *, handoff: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "POST", "/v1/a2a/handoffs/forward", tenant_id=tenant_id or self.tenant_id, body=handoff
        )

    async def forward_handoff_async(self, *, handoff: dict[str, Any], tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "POST",
            "/v1/a2a/handoffs/forward",
            tenant_id=tenant_id or self.tenant_id,
            body=handoff,
        )

    def get_receipt(self, handoff_id: str, *, tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "GET", f"/v1/a2a/handoffs/{handoff_id}/receipt", tenant_id=tenant_id or self.tenant_id
        )

    async def get_receipt_async(self, handoff_id: str, *, tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "GET",
            f"/v1/a2a/handoffs/{handoff_id}/receipt",
            tenant_id=tenant_id or self.tenant_id,
        )

    def get_lineage(self, handoff_id: str, *, tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "GET", f"/v1/a2a/handoffs/{handoff_id}/lineage", tenant_id=tenant_id or self.tenant_id
        )

    async def get_lineage_async(self, handoff_id: str, *, tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "GET",
            f"/v1/a2a/handoffs/{handoff_id}/lineage",
            tenant_id=tenant_id or self.tenant_id,
        )

    def verify_receipt(self, handoff_id: str, *, tenant_id: str | None = None) -> dict[str, Any]:
        return request_json(
            self._core, "GET", f"/v1/a2a/handoffs/{handoff_id}/verify", tenant_id=tenant_id or self.tenant_id
        )

    async def verify_receipt_async(self, handoff_id: str, *, tenant_id: str | None = None) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "GET",
            f"/v1/a2a/handoffs/{handoff_id}/verify",
            tenant_id=tenant_id or self.tenant_id,
        )

    def list_trust_scores(
        self, *, agents: list[str] | None = None, limit: int | None = None, tenant_id: str | None = None
    ) -> dict[str, Any]:
        return request_json(
            self._core,
            "GET",
            "/v1/a2a/trust-scores",
            tenant_id=tenant_id or self.tenant_id,
            params={"agents": ",".join(agents) if agents else None, "limit": limit},
        )

    async def list_trust_scores_async(
        self, *, agents: list[str] | None = None, limit: int | None = None, tenant_id: str | None = None
    ) -> dict[str, Any]:
        return await request_json_async(
            self._core,
            "GET",
            "/v1/a2a/trust-scores",
            tenant_id=tenant_id or self.tenant_id,
            params={"agents": ",".join(agents) if agents else None, "limit": limit},
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class APIKeyRecord:
    key_id: str
    tenant_id: str
    name: str
    key_prefix: str
    scopes: list[str]
    created_at: str
    expires_at: str | None = None
    revoked_at: str | None = None

    @staticmethod
    def from_dict(d: dict[str, Any]) -> APIKeyRecord:
        return APIKeyRecord(
            key_id=str(d.get("key_id") or ""),
            tenant_id=str(d.get("tenant_id") or ""),
            name=str(d.get("name") or ""),
            key_prefix=str(d.get("key_prefix") or ""),
            scopes=[str(x) for x in (d.get("scopes") or [])],
            created_at=str(d.get("created_at") or ""),
            expires_at=(str(d["expires_at"]) if d.get("expires_at") is not None else None),
            revoked_at=(str(d["revoked_at"]) if d.get("revoked_at") is not None else None),
        )



from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, ConfigDict

CostPricingSource = Literal["reported", "estimated", "mixed", "unknown"]
CostEventKind = Literal["llm", "tool", "retry", "replay", "export", "blocked", "approval_wait", "other"]


class TraceCostProviderModelBreakdown(BaseModel):
    model_config = ConfigDict(extra="allow")

    provider: str
    model: str
    usd: float | None = None
    tokens: int | None = None


class TraceCostRetryBreakdown(BaseModel):
    model_config = ConfigDict(extra="allow")

    retry_index: int
    usd: float | None = None
    tokens: int | None = None


class TraceCostToolBreakdown(BaseModel):
    model_config = ConfigDict(extra="allow")

    tool: str
    usd: float | None = None
    tokens: int | None = None


class TraceCostActorBreakdown(BaseModel):
    model_config = ConfigDict(extra="allow")

    actor: str
    usd: float | None = None
    tokens: int | None = None


class TraceCostBreakdown(BaseModel):
    model_config = ConfigDict(extra="allow")

    total_usd: float | None = None
    reported_total_usd: float | None = None
    estimated_total_usd: float | None = None
    pricing_source: CostPricingSource | None = None
    provider: str | None = None
    model: str | None = None
    input_tokens: int | None = None
    output_tokens: int | None = None
    cached_tokens: int | None = None
    total_tokens: int | None = None
    retry_count: int | None = None
    tool_call_count: int | None = None
    blocked_cost_avoided_usd: float | None = None
    blocked_token_avoided: int | None = None
    by_category: dict[str, float] | None = None
    by_provider_model: list[TraceCostProviderModelBreakdown] | None = None
    by_retry: list[TraceCostRetryBreakdown] | None = None
    by_tool: list[TraceCostToolBreakdown] | None = None
    by_actor: list[TraceCostActorBreakdown] | None = None


class TraceCostEvent(BaseModel):
    model_config = ConfigDict(extra="allow")

    event_id: str
    kind: CostEventKind
    label: str
    provider: str | None = None
    model: str | None = None
    tool: str | None = None
    actor: str | None = None
    retry_index: int | None = None
    input_tokens: int | None = None
    output_tokens: int | None = None
    cached_tokens: int | None = None
    total_tokens: int | None = None
    reported_usd: float | None = None
    estimated_usd: float | None = None
    pricing_source: CostPricingSource | None = None
    avoided_usd: float | None = None
    avoided_tokens: int | None = None
    started_at: str | None = None
    finished_at: str | None = None
    metadata: dict[str, Any] | None = None


class KernelCostSummaryRow(BaseModel):
    model_config = ConfigDict(extra="allow")

    label: str
    total_usd: float | None = None
    reported_total_usd: float | None = None
    estimated_total_usd: float | None = None
    pricing_source: CostPricingSource | None = None
    total_tokens: int | None = None
    blocked_cost_avoided_usd: float | None = None
    blocked_token_avoided: int | None = None
    trace_count: int | None = None


class KernelCostSummaryResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    tenant_id: str
    from_: str | None = None
    to: str | None = None
    group_by: Literal["session", "actor", "agent", "provider_model", "tool", "day"]
    rows: list[KernelCostSummaryRow]

    @classmethod
    def model_validate(cls, obj: Any, *args: Any, **kwargs: Any):
        if isinstance(obj, dict) and "from" in obj and "from_" not in obj:
            obj = dict(obj)
            obj["from_"] = obj.pop("from")
        return super().model_validate(obj, *args, **kwargs)


class KernelCostAvoidedRow(BaseModel):
    model_config = ConfigDict(extra="allow")

    label: str
    total_usd: float | None = None
    total_tokens: int | None = None
    event_count: int | None = None


class KernelCostAvoidedResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    tenant_id: str
    from_: str | None = None
    to: str | None = None
    rows: list[KernelCostAvoidedRow]

    @classmethod
    def model_validate(cls, obj: Any, *args: Any, **kwargs: Any):
        if isinstance(obj, dict) and "from" in obj and "from_" not in obj:
            obj = dict(obj)
            obj["from_"] = obj.pop("from")
        return super().model_validate(obj, *args, **kwargs)


class KernelCostEventsResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    tenant_id: str
    from_: str | None = None
    to: str | None = None
    events: list[TraceCostEvent]

    @classmethod
    def model_validate(cls, obj: Any, *args: Any, **kwargs: Any):
        if isinstance(obj, dict) and "from" in obj and "from_" not in obj:
            obj = dict(obj)
            obj["from_"] = obj.pop("from")
        return super().model_validate(obj, *args, **kwargs)


class KernelCostAnomaly(BaseModel):
    model_config = ConfigDict(extra="allow")

    anomaly_id: str
    kind: str
    severity: Literal["low", "medium", "high", "critical"]
    dimension: str
    baseline: float
    observed: float
    delta: float
    trace_id: str | None = None
    session_id: str | None = None
    actor: str | None = None
    agent_id: str | None = None
    provider: str | None = None
    model: str | None = None
    detected_at: str


class KernelCostAnomaliesResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    tenant_id: str
    from_: str | None = None
    to: str | None = None
    anomalies: list[KernelCostAnomaly]

    @classmethod
    def model_validate(cls, obj: Any, *args: Any, **kwargs: Any):
        if isinstance(obj, dict) and "from" in obj and "from_" not in obj:
            obj = dict(obj)
            obj["from_"] = obj.pop("from")
        return super().model_validate(obj, *args, **kwargs)



from __future__ import annotations

import json
from typing import Any

from sentinos.auth import workforce_cli


class _FakeHTTPResponse:
    def __init__(self, payload: dict[str, Any]):
        self._payload = payload

    def read(self) -> bytes:
        return json.dumps(self._payload).encode("utf-8")

    def __enter__(self) -> _FakeHTTPResponse:
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        return None


def test_cli_exchange_posts_expected_payload(monkeypatch, capsys) -> None:
    seen: dict[str, Any] = {}

    def fake_urlopen(req, timeout=15.0):  # noqa: ARG001
        seen["url"] = req.full_url
        seen["payload"] = json.loads(req.data.decode("utf-8"))
        return _FakeHTTPResponse({"tokens": {"access_token": "a1", "refresh_token": "r1"}})

    monkeypatch.setattr("urllib.request.urlopen", fake_urlopen)

    rc = workforce_cli.main(
        [
            "exchange",
            "--controlplane-url",
            "https://cp.example",
            "--org-id",
            "org-1",
            "--idp-issuer",
            "https://idp.example",
            "--external-subject",
            "emp-1",
            "--groups",
            "AI_USERS,FINANCE",
            "--audience",
            "sentinos-workforce",
        ]
    )
    assert rc == 0
    assert seen["url"] == "https://cp.example/v1/workforce/token/exchange"
    assert seen["payload"]["org_id"] == "org-1"
    assert seen["payload"]["groups"] == ["AI_USERS", "FINANCE"]
    out = capsys.readouterr().out
    assert '"access_token": "a1"' in out


def test_cli_refresh_posts_refresh_token(monkeypatch, capsys) -> None:
    seen: dict[str, Any] = {}

    def fake_urlopen(req, timeout=15.0):  # noqa: ARG001
        seen["url"] = req.full_url
        seen["payload"] = json.loads(req.data.decode("utf-8"))
        return _FakeHTTPResponse({"access_token": "a2"})

    monkeypatch.setattr("urllib.request.urlopen", fake_urlopen)

    rc = workforce_cli.main(
        [
            "refresh",
            "--controlplane-url",
            "https://cp.example",
            "--refresh-token",
            "refresh-123",
        ]
    )
    assert rc == 0
    assert seen["url"] == "https://cp.example/v1/workforce/token/refresh"
    assert seen["payload"] == {"refresh_token": "refresh-123"}
    out = capsys.readouterr().out
    assert '"access_token": "a2"' in out



"""OpenAI-style completion helper with Sentinos policy guard."""

from __future__ import annotations

from typing import Any

from sentinos import LLMGuard, SentinosClient, guard_openai_chat_completion


def fake_openai_chat_create(*, model: str, messages: list[dict[str, Any]], temperature: float = 0.2) -> dict[str, Any]:
    return {
        "id": "chatcmpl-1",
        "model": model,
        "choices": [{"index": 0, "message": {"role": "assistant", "content": "Acknowledged."}}],
        "usage": {"prompt_tokens": 10, "completion_tokens": 4},
        "temperature": temperature,
        "messages_count": len(messages),
    }


def run_openai_example() -> None:
    client = SentinosClient.from_env(org_id="acme")
    guard = LLMGuard(kernel=client.kernel, agent_id="ops-bot", session_id="sess-002")
    result = guard_openai_chat_completion(
        guard=guard,
        create=fake_openai_chat_create,
        model="gpt-4o-mini",
        messages=[
            {"role": "system", "content": "You are a SOC analyst assistant."},
            {"role": "user", "content": "Classify this alert."},
        ],
        temperature=0.1,
    )
    print(result.trace.trace_id, result.response["id"])


if __name__ == "__main__":
    run_openai_example()



"""Chronos observability analysis example."""

from sentinos import SentinosClient


def run_observability_queries(client: SentinosClient) -> dict:
    traces = client.chronos.observability_traces(limit=200)
    anomalies = client.chronos.observability_anomalies(threshold=0.8)
    patterns = client.chronos.observability_patterns()
    return {
        "traces": traces,
        "anomalies": anomalies,
        "patterns": patterns,
    }



from __future__ import annotations

import contextlib
import importlib
import json
import os
from collections.abc import AsyncIterable, AsyncIterator, Awaitable, Callable, Iterable, Iterator
from dataclasses import dataclass
from itertools import chain
from typing import Any, TypeVar

from ..models.decision_trace import DecisionTrace
from ..models.openresponses import OpenResponsesRequest, OpenResponsesResponse, OpenResponsesStreamEvent
from .llm import LLMGuard, LLMPolicyDeniedError, LLMPolicyEscalationError, LLMPolicyResult

T = TypeVar("T")


def _import_optional(module_name: str, extra_name: str) -> Any:
    try:
        return importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        raise ImportError(
            f"Optional dependency '{module_name}' is not installed. Install with: pip install 'sentinos[{extra_name}]'"
        ) from exc


def _require_attr(obj: Any, path: str) -> Any:
    cur = obj
    for part in path.split("."):
        if not hasattr(cur, part):
            raise AttributeError(f"Expected attribute '{path}' was not found on {type(obj).__name__}")
        cur = getattr(cur, part)
    return cur


def _resolve_openrouter_api_key(api_key: str | None) -> str | None:
    if api_key is not None:
        return api_key
    env_key = os.getenv("OPENROUTER_API_KEY")
    if env_key is None:
        return None
    value = env_key.strip()
    return value if value else None


def _openrouter_default_headers(
    *,
    http_referer: str | None = None,
    x_title: str | None = None,
) -> dict[str, str]:
    headers: dict[str, str] = {}
    if http_referer is not None and http_referer.strip():
        headers["HTTP-Referer"] = http_referer.strip()
    if x_title is not None and x_title.strip():
        headers["X-Title"] = x_title.strip()
    return headers


def _decision_text(trace: DecisionTrace) -> str:
    return str(trace.decision).strip().upper()


def _ensure_allowed(*, provider: str, operation: str, trace: DecisionTrace) -> str:
    decision = _decision_text(trace)
    reason = trace.policy_evaluation.reason or "policy decision"
    if decision == "DENY":
        raise LLMPolicyDeniedError(f"Sentinos denied {provider}.{operation}: {reason}", trace=trace)
    if decision == "ESCALATE":
        raise LLMPolicyEscalationError(f"Sentinos escalated {provider}.{operation}: {reason}", trace=trace)
    return decision


def _to_payload(value: Any) -> dict[str, Any]:
    if isinstance(value, dict):
        return value
    if hasattr(value, "model_dump") and callable(getattr(value, "model_dump")):
        model_dump = getattr(value, "model_dump")
        return model_dump()
    if hasattr(value, "to_dict") and callable(getattr(value, "to_dict")):
        to_dict = getattr(value, "to_dict")
        return to_dict()
    if hasattr(value, "dict") and callable(getattr(value, "dict")):
        dict_fn = getattr(value, "dict")
        return dict_fn()
    if hasattr(value, "__dict__"):
        raw = vars(value)
        return {str(k): v for k, v in raw.items() if not str(k).startswith("_")}
    raise TypeError(f"Unable to normalize payload from type {type(value).__name__}")


def _normalize_response(value: Any) -> OpenResponsesResponse:
    if isinstance(value, OpenResponsesResponse):
        return value
    return OpenResponsesResponse.model_validate(_to_payload(value))


def _summarize_response(response: OpenResponsesResponse) -> dict[str, Any]:
    summary: dict[str, Any] = {
        "id": response.id,
        "status": response.status,
        "model": response.model,
        "output_items": len(response.output or []),
    }
    if response.usage is not None:
        summary["usage"] = response.usage.model_dump(exclude_none=True)
    return summary


def parse_openresponses_sse_event(
    raw: OpenResponsesStreamEvent | dict[str, Any] | str | bytes,
    *,
    event_name: str | None = None,
) -> OpenResponsesStreamEvent | None:
    if isinstance(raw, OpenResponsesStreamEvent):
        return raw

    payload: dict[str, Any]
    if isinstance(raw, dict):
        payload = raw
    else:
        text = raw.decode("utf-8", errors="replace") if isinstance(raw, bytes) else raw
        line = text.strip()
        if not line:
            return None
        if line.startswith("data:"):
            line = line[5:].strip()
        if line == "[DONE]":
            return OpenResponsesStreamEvent(type="response.done")
        loaded = json.loads(line)
        if not isinstance(loaded, dict):
            raise ValueError("Open Responses SSE data payload must decode to an object")
        payload = loaded

    if event_name and "type" not in payload:
        payload = dict(payload)
        payload["type"] = event_name
    return OpenResponsesStreamEvent.model_validate(payload)


def iter_openresponses_sse_lines(lines: Iterable[str]) -> Iterator[OpenResponsesStreamEvent]:
    event_name: str | None = None
    data_lines: list[str] = []

    def flush() -> OpenResponsesStreamEvent | None:
        nonlocal event_name, data_lines
        if event_name is None and not data_lines:
            return None
        payload = "\n".join(data_lines).strip()
        out = parse_openresponses_sse_event(payload if payload else "{}", event_name=event_name)
        event_name = None
        data_lines = []
        return out

    for raw_line in lines:
        line = raw_line.rstrip("\n")
        if line == "":
            event = flush()
            if event is not None:
                yield event
            continue
        if line.startswith(":"):
            continue
        if line.startswith("event:"):
            event_name = line[6:].strip()
            continue
        if line.startswith("data:"):
            data_lines.append(line[5:].strip())
            continue
        # Accept JSON-lines style streams where lines are full payloads.
        data_lines.append(line.strip())

    tail = flush()
    if tail is not None:
        yield tail


async def aiter_openresponses_sse_lines(lines: AsyncIterable[str]) -> AsyncIterator[OpenResponsesStreamEvent]:
    event_name: str | None = None
    data_lines: list[str] = []

    async def flush() -> OpenResponsesStreamEvent | None:
        nonlocal event_name, data_lines
        if event_name is None and not data_lines:
            return None
        payload = "\n".join(data_lines).strip()
        out = parse_openresponses_sse_event(payload if payload else "{}", event_name=event_name)
        event_name = None
        data_lines = []
        return out

    async for raw_line in lines:
        line = raw_line.rstrip("\n")
        if line == "":
            event = await flush()
            if event is not None:
                yield event
            continue
        if line.startswith(":"):
            continue
        if line.startswith("event:"):
            event_name = line[6:].strip()
            continue
        if line.startswith("data:"):
            data_lines.append(line[5:].strip())
            continue
        data_lines.append(line.strip())

    tail = await flush()
    if tail is not None:
        yield tail


def _iter_items(stream_obj: Any) -> Iterator[Any]:
    if hasattr(stream_obj, "events") and callable(getattr(stream_obj, "events")):
        yield from stream_obj.events()
        return
    if hasattr(stream_obj, "iter_lines") and callable(getattr(stream_obj, "iter_lines")):
        yield from stream_obj.iter_lines()
        return
    if isinstance(stream_obj, Iterable) and not isinstance(stream_obj, (str, bytes, dict)):
        yield from stream_obj
        return
    raise TypeError(f"Unsupported stream object type: {type(stream_obj).__name__}")


async def _aiter_items(stream_obj: Any) -> AsyncIterator[Any]:
    if hasattr(stream_obj, "__aiter__"):
        async for item in stream_obj:
            yield item
        return
    if hasattr(stream_obj, "aiter_lines") and callable(getattr(stream_obj, "aiter_lines")):
        async for line in stream_obj.aiter_lines():
            yield line
        return
    if isinstance(stream_obj, Iterable) and not isinstance(stream_obj, (str, bytes, dict)):
        for item in stream_obj:
            yield item
        return
    raise TypeError(f"Unsupported async stream object type: {type(stream_obj).__name__}")


def _extract_final_response(stream_obj: Any) -> OpenResponsesResponse | None:
    getter = getattr(stream_obj, "get_final_response", None)
    if getter is None or not callable(getter):
        return None
    with contextlib.suppress(Exception):
        return _normalize_response(getter())
    return None


@dataclass(frozen=True)
class OpenResponsesStreamResult:
    provider: str
    operation: str
    decision: str
    trace: DecisionTrace
    events: list[OpenResponsesStreamEvent]
    final_response: OpenResponsesResponse | None = None


@dataclass
class OpenResponsesAdapter:
    """
    Governance adapter for Open Responses-style `responses.create` integrations.

    Supports:
    - OpenAI Responses compatible clients
    - provider/router endpoints implementing the Open Responses schema
    - custom callables via `create_fn` / `stream_fn`
    """

    guard: LLMGuard
    create_fn: Callable[..., Any]
    stream_fn: Callable[..., Any] | None = None
    provider: str = "openresponses"
    operation: str = "responses.create"
    create_uses_body: bool = False
    stream_uses_body: bool = False

    @classmethod
    def from_client(cls, *, guard: LLMGuard, client: Any) -> OpenResponsesAdapter:
        create_fn = _require_attr(client, "responses.create")
        stream_fn = None
        with contextlib.suppress(AttributeError):
            candidate = _require_attr(client, "responses.stream")
            if callable(candidate):
                stream_fn = candidate
        return cls(guard=guard, create_fn=create_fn, stream_fn=stream_fn)

    def _build_body(self, *, model: str, input: Any, kwargs: dict[str, Any]) -> dict[str, Any]:
        request = OpenResponsesRequest.model_validate({"model": model, "input": input, **kwargs})
        return request.model_dump(exclude_none=True)

    def _invoke_create(self, body: dict[str, Any]) -> OpenResponsesResponse:
        raw = self.create_fn(body) if self.create_uses_body else self.create_fn(**body)
        return _normalize_response(raw)

    def create(
        self,
        *,
        model: str,
        input: Any,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> LLMPolicyResult[OpenResponsesResponse]:
        body = self._build_body(model=model, input=input, kwargs=kwargs)
        return self.guard.run(
            provider=self.provider,
            operation=self.operation,
            request=body,
            model=model,
            metadata=metadata,
            tenant_id=tenant_id,
            invoke=lambda: self._invoke_create(body),
            response_summarizer=_summarize_response,
        )

    def create_request(
        self,
        *,
        request: OpenResponsesRequest | dict[str, Any],
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
    ) -> LLMPolicyResult[OpenResponsesResponse]:
        body = request.model_dump(exclude_none=True) if isinstance(request, OpenResponsesRequest) else dict(request)
        model_raw = body.get("model")
        model = model_raw if isinstance(model_raw, str) else None
        return self.guard.run(
            provider=self.provider,
            operation=self.operation,
            request=body,
            model=model,
            metadata=metadata,
            tenant_id=tenant_id,
            invoke=lambda: self._invoke_create(body),
            response_summarizer=_summarize_response,
        )

    def _collect_stream(self, stream_obj: Any) -> tuple[list[OpenResponsesStreamEvent], OpenResponsesResponse | None]:
        events: list[OpenResponsesStreamEvent] = []
        final_response: OpenResponsesResponse | None = None

        iterator = _iter_items(stream_obj)
        try:
            first = next(iterator)
        except StopIteration:
            return events, _extract_final_response(stream_obj)

        joined = chain([first], iterator)
        if isinstance(first, (str, bytes)):
            normalized_lines = (
                line.decode("utf-8", errors="replace") if isinstance(line, bytes) else line for line in joined
            )
            for event in iter_openresponses_sse_lines(normalized_lines):
                events.append(event)
                if event.response is not None:
                    final_response = event.response
        else:
            for raw in joined:
                event = OpenResponsesStreamEvent.model_validate(_to_payload(raw))
                events.append(event)
                if event.response is not None:
                    final_response = event.response

        if final_response is None:
            final_response = _extract_final_response(stream_obj)
        return events, final_response

    def stream(
        self,
        *,
        model: str,
        input: Any,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> OpenResponsesStreamResult:
        if self.stream_fn is None:
            raise ValueError("stream_fn is not configured; provide one or use a client exposing responses.stream")

        body = self._build_body(model=model, input=input, kwargs={**kwargs, "stream": True})
        trace = self.guard.authorize(
            provider=self.provider,
            operation=self.operation,
            request=body,
            model=model,
            metadata=metadata,
            tenant_id=tenant_id,
        )
        decision = _ensure_allowed(provider=self.provider, operation=self.operation, trace=trace)

        raw_stream = self.stream_fn(body) if self.stream_uses_body else self.stream_fn(**body)
        if hasattr(raw_stream, "__enter__") and hasattr(raw_stream, "__exit__"):
            with raw_stream as active_stream:
                events, final_response = self._collect_stream(active_stream)
        else:
            events, final_response = self._collect_stream(raw_stream)

        if final_response is not None:
            with contextlib.suppress(Exception):
                self.guard.record_response(
                    provider=self.provider,
                    operation=self.operation,
                    trace=trace,
                    response=final_response,
                    tenant_id=tenant_id,
                    response_summarizer=lambda r: _summarize_response(_normalize_response(r)),
                )
        return OpenResponsesStreamResult(
            provider=self.provider,
            operation=self.operation,
            decision=decision,
            trace=trace,
            events=events,
            final_response=final_response,
        )


@dataclass
class AsyncOpenResponsesAdapter:
    guard: LLMGuard
    create_fn: Callable[..., Awaitable[Any]]
    stream_fn: Callable[..., Awaitable[Any] | Any] | None = None
    provider: str = "openresponses"
    operation: str = "responses.create"
    create_uses_body: bool = False
    stream_uses_body: bool = False

    @classmethod
    def from_client(cls, *, guard: LLMGuard, client: Any) -> AsyncOpenResponsesAdapter:
        create_fn = _require_attr(client, "responses.create")
        stream_fn = None
        with contextlib.suppress(AttributeError):
            candidate = _require_attr(client, "responses.stream")
            if callable(candidate):
                stream_fn = candidate
        return cls(guard=guard, create_fn=create_fn, stream_fn=stream_fn)

    def _build_body(self, *, model: str, input: Any, kwargs: dict[str, Any]) -> dict[str, Any]:
        request = OpenResponsesRequest.model_validate({"model": model, "input": input, **kwargs})
        return request.model_dump(exclude_none=True)

    async def _invoke_create(self, body: dict[str, Any]) -> OpenResponsesResponse:
        raw = await (self.create_fn(body) if self.create_uses_body else self.create_fn(**body))
        return _normalize_response(raw)

    async def create(
        self,
        *,
        model: str,
        input: Any,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> LLMPolicyResult[OpenResponsesResponse]:
        body = self._build_body(model=model, input=input, kwargs=kwargs)
        return await self.guard.run_async(
            provider=self.provider,
            operation=self.operation,
            request=body,
            model=model,
            metadata=metadata,
            tenant_id=tenant_id,
            invoke=lambda: self._invoke_create(body),
            response_summarizer=_summarize_response,
        )

    async def create_request(
        self,
        *,
        request: OpenResponsesRequest | dict[str, Any],
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
    ) -> LLMPolicyResult[OpenResponsesResponse]:
        body = request.model_dump(exclude_none=True) if isinstance(request, OpenResponsesRequest) else dict(request)
        model_raw = body.get("model")
        model = model_raw if isinstance(model_raw, str) else None
        return await self.guard.run_async(
            provider=self.provider,
            operation=self.operation,
            request=body,
            model=model,
            metadata=metadata,
            tenant_id=tenant_id,
            invoke=lambda: self._invoke_create(body),
            response_summarizer=_summarize_response,
        )

    async def _collect_stream(
        self,
        stream_obj: Any,
    ) -> tuple[list[OpenResponsesStreamEvent], OpenResponsesResponse | None]:
        events: list[OpenResponsesStreamEvent] = []
        final_response: OpenResponsesResponse | None = None

        iterator = _aiter_items(stream_obj)
        try:
            first = await iterator.__anext__()
        except StopAsyncIteration:
            return events, _extract_final_response(stream_obj)

        if isinstance(first, (str, bytes)):

            async def line_iter() -> AsyncIterator[str]:
                first_line = first.decode("utf-8", errors="replace") if isinstance(first, bytes) else first
                yield first_line
                async for line in iterator:
                    yield line.decode("utf-8", errors="replace") if isinstance(line, bytes) else line

            async for event in aiter_openresponses_sse_lines(line_iter()):
                events.append(event)
                if event.response is not None:
                    final_response = event.response
        else:
            first_event = OpenResponsesStreamEvent.model_validate(_to_payload(first))
            events.append(first_event)
            if first_event.response is not None:
                final_response = first_event.response
            async for raw in iterator:
                event = OpenResponsesStreamEvent.model_validate(_to_payload(raw))
                events.append(event)
                if event.response is not None:
                    final_response = event.response

        if final_response is None:
            final_response = _extract_final_response(stream_obj)
        return events, final_response

    async def stream(
        self,
        *,
        model: str,
        input: Any,
        metadata: dict[str, Any] | None = None,
        tenant_id: str | None = None,
        **kwargs: Any,
    ) -> OpenResponsesStreamResult:
        if self.stream_fn is None:
            raise ValueError("stream_fn is not configured; provide one or use a client exposing responses.stream")

        body = self._build_body(model=model, input=input, kwargs={**kwargs, "stream": True})
        trace = await self.guard.authorize_async(
            provider=self.provider,
            operation=self.operation,
            request=body,
            model=model,
            metadata=metadata,
            tenant_id=tenant_id,
        )
        decision = _ensure_allowed(provider=self.provider, operation=self.operation, trace=trace)

        candidate = self.stream_fn(body) if self.stream_uses_body else self.stream_fn(**body)
        raw_stream = await candidate if isinstance(candidate, Awaitable) else candidate

        if hasattr(raw_stream, "__aenter__") and hasattr(raw_stream, "__aexit__"):
            async with raw_stream as active_stream:
                events, final_response = await self._collect_stream(active_stream)
        else:
            events, final_response = await self._collect_stream(raw_stream)

        if final_response is not None:
            with contextlib.suppress(Exception):
                await self.guard.record_response_async(
                    provider=self.provider,
                    operation=self.operation,
                    trace=trace,
                    response=final_response,
                    tenant_id=tenant_id,
                    response_summarizer=lambda r: _summarize_response(_normalize_response(r)),
                )
        return OpenResponsesStreamResult(
            provider=self.provider,
            operation=self.operation,
            decision=decision,
            trace=trace,
            events=events,
            final_response=final_response,
        )


def create_openresponses_adapter(
    *,
    guard: LLMGuard,
    client: Any | None = None,
    create_fn: Callable[..., Any] | None = None,
    stream_fn: Callable[..., Any] | None = None,
    api_key: str | None = None,
    base_url: str | None = None,
    organization: str | None = None,
    default_headers: dict[str, str] | None = None,
    provider: str = "openresponses",
    operation: str = "responses.create",
) -> OpenResponsesAdapter:
    if create_fn is not None:
        return OpenResponsesAdapter(
            guard=guard,
            create_fn=create_fn,
            stream_fn=stream_fn,
            provider=provider,
            operation=operation,
        )

    c = client
    if c is None:
        openai_module = _import_optional("openai", "openai")
        ctor = _require_attr(openai_module, "OpenAI")
        kwargs: dict[str, Any] = {}
        if api_key is not None:
            kwargs["api_key"] = api_key
        if base_url is not None:
            kwargs["base_url"] = base_url
        if organization is not None:
            kwargs["organization"] = organization
        if default_headers:
            kwargs["default_headers"] = default_headers
        c = ctor(**kwargs)
    adapter = OpenResponsesAdapter.from_client(guard=guard, client=c)
    adapter.provider = provider
    adapter.operation = operation
    if stream_fn is not None:
        adapter.stream_fn = stream_fn
    return adapter


def create_async_openresponses_adapter(
    *,
    guard: LLMGuard,
    client: Any | None = None,
    create_fn: Callable[..., Awaitable[Any]] | None = None,
    stream_fn: Callable[..., Awaitable[Any] | Any] | None = None,
    api_key: str | None = None,
    base_url: str | None = None,
    organization: str | None = None,
    default_headers: dict[str, str] | None = None,
    provider: str = "openresponses",
    operation: str = "responses.create",
) -> AsyncOpenResponsesAdapter:
    if create_fn is not None:
        return AsyncOpenResponsesAdapter(
            guard=guard,
            create_fn=create_fn,
            stream_fn=stream_fn,
            provider=provider,
            operation=operation,
        )

    c = client
    if c is None:
        openai_module = _import_optional("openai", "openai")
        ctor = _require_attr(openai_module, "AsyncOpenAI")
        kwargs: dict[str, Any] = {}
        if api_key is not None:
            kwargs["api_key"] = api_key
        if base_url is not None:
            kwargs["base_url"] = base_url
        if organization is not None:
            kwargs["organization"] = organization
        if default_headers:
            kwargs["default_headers"] = default_headers
        c = ctor(**kwargs)
    adapter = AsyncOpenResponsesAdapter.from_client(guard=guard, client=c)
    adapter.provider = provider
    adapter.operation = operation
    if stream_fn is not None:
        adapter.stream_fn = stream_fn
    return adapter


def create_openrouter_openresponses_adapter(
    *,
    guard: LLMGuard,
    client: Any | None = None,
    create_fn: Callable[..., Any] | None = None,
    stream_fn: Callable[..., Any] | None = None,
    api_key: str | None = None,
    base_url: str = "https://openrouter.ai/api/v1",
    organization: str | None = None,
    http_referer: str | None = None,
    x_title: str | None = None,
) -> OpenResponsesAdapter:
    resolved_api_key = _resolve_openrouter_api_key(api_key)
    return create_openresponses_adapter(
        guard=guard,
        client=client,
        create_fn=create_fn,
        stream_fn=stream_fn,
        api_key=resolved_api_key,
        base_url=base_url,
        organization=organization,
        default_headers=_openrouter_default_headers(http_referer=http_referer, x_title=x_title),
        provider="openrouter",
        operation="responses.create",
    )


def create_async_openrouter_openresponses_adapter(
    *,
    guard: LLMGuard,
    client: Any | None = None,
    create_fn: Callable[..., Awaitable[Any]] | None = None,
    stream_fn: Callable[..., Awaitable[Any] | Any] | None = None,
    api_key: str | None = None,
    base_url: str = "https://openrouter.ai/api/v1",
    organization: str | None = None,
    http_referer: str | None = None,
    x_title: str | None = None,
) -> AsyncOpenResponsesAdapter:
    resolved_api_key = _resolve_openrouter_api_key(api_key)
    return create_async_openresponses_adapter(
        guard=guard,
        client=client,
        create_fn=create_fn,
        stream_fn=stream_fn,
        api_key=resolved_api_key,
        base_url=base_url,
        organization=organization,
        default_headers=_openrouter_default_headers(http_referer=http_referer, x_title=x_title),
        provider="openrouter",
        operation="responses.create",
    )


def guard_openresponses_create(
    *,
    guard: LLMGuard,
    create: Callable[..., Any],
    model: str,
    input: Any,
    metadata: dict[str, Any] | None = None,
    tenant_id: str | None = None,
    **kwargs: Any,
) -> LLMPolicyResult[OpenResponsesResponse]:
    adapter = OpenResponsesAdapter(guard=guard, create_fn=create)
    return adapter.create(model=model, input=input, metadata=metadata, tenant_id=tenant_id, **kwargs)



from .circuit_breaker import CircuitBreaker, CircuitOpenError
from .retry import RetryPolicy
from .telemetry import TelemetryConfig, httpx_event_hooks

__all__ = (
    "CircuitBreaker",
    "CircuitOpenError",
    "RetryPolicy",
    "TelemetryConfig",
    "httpx_event_hooks",
)
